from __future__ import annotations

from typing import Any, Literal

from x8.content.file import FileData
from x8.content.image import ImageData
from x8.core import DataModel


class Function(DataModel):
    type: Literal["function"] = "function"
    name: str
    description: str | None = None
    parameters: dict[str, Any] | None = None
    strict: bool | None = None


class WebSearchFilters(DataModel):
    allowed_domains: list[str] | None = None


class WebSearchUserLocation(DataModel):
    city: str | None = None
    country: str | None = None
    region: str | None = None
    timezone: str | None = None
    type: Literal["approximate"] = "approximate"


class WebSearch(DataModel):
    type: Literal["web_search"] = "web_search"
    search_context_size: Literal["short", "medium", "long"] = "medium"
    filters: WebSearchFilters | None = None
    user_location: WebSearchUserLocation | None = None


Tool = Function | WebSearch


class ToolChoiceFunction(DataModel):
    type: Literal["function"] = "function"
    name: str


class AllowedTools(DataModel):
    type: Literal["allowed_tools"] = "allowed_tools"
    mode: Literal["none", "auto", "required"] = "auto"
    tools: list[ToolChoiceFunction] | None = None


ToolChoice = (
    Literal["none", "auto", "required"] | ToolChoiceFunction | AllowedTools
)


class InputImage(DataModel):
    type: Literal["input_image"] = "input_image"
    detail: Literal["low", "high", "auto"] = "auto"
    image: str | ImageData


class InputText(DataModel):
    type: Literal["input_text"] = "input_text"
    text: str


class InputFile(DataModel):
    type: Literal["input_file"] = "input_file"
    file: str | FileData


InputMessageContent = InputText | InputImage | InputFile


class InputMessage(DataModel):
    type: Literal["message"] = "message"
    role: Literal["user", "system", "assistant", "developer"] = "user"
    content: str | list[InputMessageContent]


class FunctionCall(DataModel):
    type: Literal["function_call"] = "function_call"
    call_id: str | None = None
    name: str
    id: str | None = None
    arguments: str | dict[str, Any] | None = None
    status: Literal["completed", "in_progress", "incomplete"] | None = None


class FunctionCallOutput(DataModel):
    type: Literal["function_call_output"] = "function_call_output"
    call_id: str | None = None
    output: str | dict[str, Any] | None = None
    id: str | None = None
    status: Literal["completed", "in_progress", "incomplete"] | None = None


class OutputText(DataModel):
    type: Literal["output_text"] = "output_text"
    text: str
    annotations: list[dict[str, Any]] | None = None


class Refusal(DataModel):
    type: Literal["refusal"] = "refusal"
    refusal: str


class OutputReasoningSummaryText(DataModel):
    type: Literal["summary_text"] = "summary_text"
    text: str


class OutputReasoningContentText(DataModel):
    type: Literal["reasoning_text"] = "reasoning_text"
    text: str


class OutputReasoning(DataModel):
    id: str | None = None
    type: Literal["reasoning"] = "reasoning"
    summary: list[OutputReasoningSummaryText] | None = None
    content: list[OutputReasoningContentText] | None = None
    status: Literal["completed", "in_progress", "incomplete"] | None = (
        "completed"
    )


OutputMessageContent = OutputText | Refusal | OutputReasoningContentText


class OutputMessage(DataModel):
    type: Literal["message"] = "message"
    role: Literal["assistant"] = "assistant"
    status: Literal["completed", "in_progress", "incomplete"] | None = (
        "completed"
    )
    content: list[OutputMessageContent]
    id: str | None = None


InputItem = (
    InputMessage
    | OutputMessage
    | FunctionCall
    | FunctionCallOutput
    | OutputReasoning
)
OutputItem = OutputMessage | FunctionCall | OutputReasoning


class ErrorDetail(DataModel):
    code: str | None = None
    message: str | None = None


class Usage(DataModel):
    input_tokens: int
    input_tokens_details: dict[str, int] | None = None
    output_tokens: int
    output_tokens_details: dict[str, int] | None = None
    total_tokens: int


class Reasoning(DataModel):
    effort: Literal["none", "low", "medium", "high"] | None = None
    summary: Literal["auto", "concise", "detailed"] | None = None


class TextGenerationResult(DataModel):
    id: str | None = None
    model: str | None = None
    created_at: int | None = None
    status: (
        Literal[
            "completed",
            "failed",
            "in_progress",
            "cancelled",
            "queued",
            "incomplete",
        ]
        | None
    ) = "completed"
    error: ErrorDetail | None = None
    output: list[OutputItem] | None = None
    usage: Usage | None = None


class StreamEvent(DataModel):
    type: str
    sequence_number: int | None = None


class ResponseQueuedEvent(StreamEvent):
    type: Literal["queued"] = "queued"
    response: TextGenerationResult


class ResponseCreatedEvent(StreamEvent):
    type: Literal["created"] = "created"
    response: TextGenerationResult


class ResponseInProgressEvent(StreamEvent):
    type: Literal["in_progress"] = "in_progress"
    response: TextGenerationResult


class ResponseCompletedEvent(StreamEvent):
    type: Literal["completed"] = "completed"
    response: TextGenerationResult


class ResponseFailedEvent(StreamEvent):
    type: Literal["failed"] = "failed"
    response: TextGenerationResult


class ResponseIncompleteEvent(StreamEvent):
    type: Literal["incomplete"] = "incomplete"
    response: TextGenerationResult


class ResponseOutputTextDeltaEvent(StreamEvent):
    type: Literal["output_text_delta"] = "output_text_delta"
    delta: str
    item_id: str | None = None
    output_index: int | None = None
    content_index: int | None = None


class ResponseOutputTextDoneEvent(StreamEvent):
    type: Literal["output_text_done"] = "output_text_done"
    text: str
    item_id: str | None = None
    output_index: int | None = None
    content_index: int | None = None


class ResponseOutputItemAddedEvent(StreamEvent):
    type: Literal["output_item_added"] = "output_item_added"
    item: OutputItem
    output_index: int | None = None


class ResponseOutputItemDoneEvent(StreamEvent):
    type: Literal["output_item_done"] = "output_item_done"
    item: OutputItem
    output_index: int | None = None


class ResponseContentPartAddedEvent(StreamEvent):
    type: Literal["content_part_added"] = "content_part_added"
    part: OutputMessageContent
    item_id: str | None = None
    output_index: int | None = None
    content_index: int | None = None


class ResponseContentPartDoneEvent(StreamEvent):
    type: Literal["content_part_done"] = "content_part_done"
    part: OutputMessageContent
    item_id: str | None = None
    output_index: int | None = None
    content_index: int | None = None


class ResponseRefusalDeltaEvent(StreamEvent):
    type: Literal["refusal_delta"] = "refusal_delta"
    delta: str
    item_id: str | None = None
    output_index: int | None = None
    content_index: int | None = None


class ResponseRefusalDoneEvent(StreamEvent):
    type: Literal["refusal_done"] = "refusal_done"
    refusal: str
    item_id: str | None = None
    output_index: int | None = None
    content_index: int | None = None


class ResponseFunctionCallArgumentsDeltaEvent(StreamEvent):
    type: Literal["function_call_arguments_delta"] = (
        "function_call_arguments_delta"
    )
    delta: str
    item_id: str | None = None
    output_index: int | None = None


class ResponseFunctionCallArgumentsDoneEvent(StreamEvent):
    type: Literal["function_call_arguments_done"] = (
        "function_call_arguments_done"
    )
    name: str
    arguments: str
    item_id: str | None = None
    output_index: int | None = None


class ResponseReasoningSummaryPartAddedEvent(StreamEvent):
    type: Literal["reasoning_summary_part_added"] = (
        "reasoning_summary_part_added"
    )
    part: OutputReasoningSummaryText
    item_id: str | None = None
    output_index: int | None = None
    summary_index: int | None = None


class ResponseReasoningSummaryPartDoneEvent(StreamEvent):
    type: Literal["reasoning_summary_part_done"] = (
        "reasoning_summary_part_done"
    )
    part: OutputReasoningSummaryText
    item_id: str | None = None
    output_index: int | None = None
    summary_index: int | None = None


class ResponseReasoningSummaryTextDeltaEvent(StreamEvent):
    type: Literal["reasoning_summary_text_delta"] = (
        "reasoning_summary_text_delta"
    )
    delta: str
    item_id: str | None = None
    output_index: int | None = None
    summary_index: int | None = None


class ResponseReasoningSummaryTextDoneEvent(StreamEvent):
    type: Literal["reasoning_summary_text_done"] = (
        "reasoning_summary_text_done"
    )
    text: str
    item_id: str | None = None
    output_index: int | None = None
    summary_index: int | None = None


class ResponseReasoningTextDeltaEvent(StreamEvent):
    type: Literal["reasoning_text_delta"] = "reasoning_text_delta"
    delta: str
    item_id: str | None = None
    content_index: int | None = None
    output_index: int | None = None


class ResponseReasoningTextDoneEvent(StreamEvent):
    type: Literal["reasoning_text_done"] = "reasoning_text_done"
    text: str
    item_id: str | None = None
    content_index: int | None = None
    output_index: int | None = None


class ErrorEvent(StreamEvent):
    type: Literal["error"] = "error"
    code: str | None = None
    message: str | None = None
    param: str | None = None


TextGenerationStreamEvent = (
    ResponseQueuedEvent
    | ResponseCreatedEvent
    | ResponseInProgressEvent
    | ResponseCompletedEvent
    | ResponseFailedEvent
    | ResponseIncompleteEvent
    | ResponseOutputTextDeltaEvent
    | ResponseOutputTextDoneEvent
    | ResponseOutputItemAddedEvent
    | ResponseOutputItemDoneEvent
    | ResponseContentPartAddedEvent
    | ResponseContentPartDoneEvent
    | ResponseRefusalDeltaEvent
    | ResponseRefusalDoneEvent
    | ResponseFunctionCallArgumentsDeltaEvent
    | ResponseFunctionCallArgumentsDoneEvent
    | ResponseReasoningSummaryPartAddedEvent
    | ResponseReasoningSummaryPartDoneEvent
    | ResponseReasoningSummaryTextDeltaEvent
    | ResponseReasoningSummaryTextDoneEvent
    | ResponseReasoningTextDeltaEvent
    | ResponseReasoningTextDoneEvent
    | ErrorEvent
)



from x8.core import Response
from x8.storage.config_store import ConfigItem, ConfigList

from ._providers import get_component


class ConfigStoreSyncAndAsyncClient:
    def __init__(self, provider_type: str, async_call: bool):
        self.client = get_component(provider_type)
        self.async_call = async_call
        self.provider_type = provider_type

    async def get(
        self,
        key,
        **kwargs,
    ) -> Response[ConfigItem]:
        if self.async_call:
            return await self.client.aget(
                key=key,
                **kwargs,
            )
        return self.client.get(
            key,
            **kwargs,
        )

    async def put(
        self,
        key,
        value,
        **kwargs,
    ) -> Response[ConfigItem]:
        if self.async_call:
            return await self.client.aput(
                key,
                value,
                **kwargs,
            )
        return self.client.put(
            key,
            value,
            **kwargs,
        )

    async def delete(
        self,
        key,
        **kwargs,
    ) -> Response[None]:
        if self.async_call:
            return await self.client.adelete(
                key,
                **kwargs,
            )
        return self.client.delete(
            key,
            **kwargs,
        )

    async def query(
        self,
        where=None,
        **kwargs,
    ) -> Response[ConfigList]:
        if self.async_call:
            return await self.client.aquery(
                where,
                **kwargs,
            )
        return self.client.query(
            where,
            **kwargs,
        )

    async def count(
        self,
        where=None,
        **kwargs,
    ) -> Response[int]:
        if self.async_call:
            return await self.client.acount(
                where,
                **kwargs,
            )
        return self.client.count(
            where,
            **kwargs,
        )

    async def close(self) -> Response[None]:
        if self.async_call:
            return await self.client.aclose()
        else:
            return Response(result=None)



from typing import Any

import httpx
from x8.core import Context, Operation, Provider, TypeConverter
from x8.core.exceptions import (
    BadRequestError,
    ConflictError,
    ForbiddenError,
    InternalError,
    NotFoundError,
    NotModified,
    NotSupportedError,
    PreconditionFailedError,
    UnauthorizedError,
)

from ._helper import OperationInfo, get_component, get_operations
from ._models import ComponentMapping


class APIProvider(Provider):
    base_url: str
    credential: str | None
    timeout: float | None
    nparams: dict[str, Any]

    _init: bool
    _component_mapping: ComponentMapping
    _operation_mappings: dict[str, OperationInfo]

    def __init__(
        self,
        base_url: str,
        credential: str | None = None,
        timeout: float | None = 60,
        nparams: dict[str, object] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            base_url:
                HTTP base URL.
            credential:
                API credential.
            timeout:
                HTTP timeout. Defaults to 60 seconds.
            nparams:
                Native params to httpx client.
        """
        self.base_url = base_url
        self.credential = credential
        self.timeout = timeout
        self.nparams = nparams
        self._init = False

    def _init_mappings(self) -> None:
        if self._init:
            return
        self._component_mapping = get_component(
            ComponentMapping(component=self.__component__)
        )
        self._operation_mappings = get_operations(self._component_mapping)
        self._init = True

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self._init_mappings()
        if not operation or not operation.name:
            # TODO: can call generic method
            raise BadRequestError("Operation name is required")

        op_info = self._get_operation_info(operation.name)
        url = self._get_url(operation, op_info)
        body = self._get_body(operation, op_info)
        headers = self._get_headers()
        with httpx.Client(timeout=self.timeout) as client:
            try:
                method = op_info.http_method.lower()
                response = client.request(
                    method, url, json=body, headers=headers
                )
                response.raise_for_status()
                content_type = response.headers.get("content-type", "").lower()
                return_type = op_info.return_type
                if "application/x-ndjson" in content_type:

                    def _iter():
                        for line in response.iter_lines():
                            if not line.strip():
                                continue
                            data = httpx.Response.json(
                                httpx.Response(200, text=line)
                            )
                            yield (
                                TypeConverter.convert_value(data, return_type)
                                if return_type
                                else data
                            )

                    return _iter()
                elif (
                    "application/octet-stream" in content_type
                    or "video/" in content_type
                    or "image/" in content_type
                    or "audio/" in content_type
                ):
                    if op_info.is_return_iterator:

                        def _iter_bytes():
                            for chunk in response.iter_bytes():
                                yield chunk

                        return _iter_bytes()
                    else:
                        # return full bytes content
                        return response.content
                elif "application/json" in content_type:
                    data = response.json()
                    if return_type:
                        return TypeConverter.convert_value(data, return_type)
                    return data
                else:
                    return response.text
            except httpx.HTTPStatusError as e:
                raise self._handle_http_error(e.response)

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self._init_mappings()
        if not operation or not operation.name:
            # TODO: can call generic method
            raise BadRequestError("Operation name is required")
        op_info = self._get_operation_info(operation.name)
        url = self._get_url(operation, op_info)
        body = self._get_body(operation, op_info)
        headers = self._get_headers()
        async with httpx.AsyncClient(timeout=self.timeout) as client:
            try:
                method = op_info.http_method.lower()
                response = await client.request(
                    method, url, json=body, headers=headers
                )
                response.raise_for_status()
                content_type = response.headers.get("content-type", "").lower()
                return_type = op_info.return_type
                if "application/x-ndjson" in content_type:

                    async def _aiter():
                        async for line in response.aiter_lines():
                            if not line.strip():
                                continue
                            data = httpx.Response.json(
                                httpx.Response(200, text=line)
                            )
                            yield (
                                TypeConverter.convert_value(data, return_type)
                                if return_type
                                else data
                            )

                    return _aiter()
                elif (
                    "application/octet-stream" in content_type
                    or "video/" in content_type
                    or "image/" in content_type
                    or "audio/" in content_type
                ):
                    if op_info.is_return_iterator:

                        async def _aiter_bytes():
                            async for chunk in response.aiter_bytes():
                                yield chunk

                        return _aiter_bytes()
                    else:
                        # return full bytes content
                        return response.content
                elif "application/json" in content_type:
                    data = response.json()
                    if return_type:
                        return TypeConverter.convert_value(data, return_type)
                    return data
                else:
                    return response.text

            except httpx.HTTPStatusError as e:
                raise self._handle_http_error(e.response)

    def _get_operation_info(self, name: str) -> OperationInfo:
        if name not in self._operation_mappings:
            raise BadRequestError(f"Operation '{name}' not found")
        return self._operation_mappings[name]

    def _get_headers(self) -> dict[str, str]:
        headers: dict[str, str] = {}
        if self.credential:
            headers["Authorization"] = f"Bearer {self.credential}"
        return headers

    def _get_url(
        self,
        operation: Operation,
        operation_info: OperationInfo,
    ) -> str:
        base = self.base_url.rstrip("/")
        path = operation_info.path.lstrip("/")
        if path:
            url = f"{base}/{path}"
        else:
            url = base
        for path_param in operation_info.path_params:
            if operation.args is None or path_param not in operation.args:
                raise BadRequestError(
                    f"Path parameter '{path_param}' is required"
                )
            value = str(operation.args[path_param])
            url = url.replace(f"{{{path_param}}}", value)
        return url

    def _get_body(
        self, operation: Operation, operation_info: OperationInfo
    ) -> dict[str, Any] | None:
        args = operation.args or {}
        for path_param in operation_info.path_params:
            args.pop(path_param, None)
        return args if args else None

    def _handle_http_error(self, response: httpx.Response) -> Exception:
        if response.status_code == 304:
            return NotModified(response.text)
        if response.status_code == 400:
            return BadRequestError(response.text)
        elif response.status_code == 401:
            return UnauthorizedError(response.text)
        elif response.status_code == 403:
            return ForbiddenError(response.text)
        elif response.status_code == 404:
            return NotFoundError(response.text)
        elif response.status_code == 409:
            return ConflictError(response.text)
        elif response.status_code == 412:
            return PreconditionFailedError(response.text)
        elif response.status_code == 415:
            return NotSupportedError(response.text)
        elif response.status_code == 500:
            return InternalError(response.text)
        return InternalError(f"{response.status_code}: {response.text}")



from typing import Any, List

from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import Containerizer
from x8.core import Component, Response, RunContext, operation
from x8.ql import Expression

from ._models import (
    Revision,
    ServiceDefinition,
    ServiceItem,
    TrafficAllocation,
)


class ContainerDeployment(Component):
    service: ServiceDefinition | None = None
    containerizer: Containerizer | None = None
    container_registry: ContainerRegistry | None = None

    def __init__(
        self,
        service: ServiceDefinition | None = None,
        containerizer: Containerizer | None = None,
        container_registry: ContainerRegistry | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            service:
                Service definition to deploy.
            containerizer:
                Containerizer to use for building the container.
            container_registry:
                Container registry to use for pushing the container.
        """
        self.service = service
        self.containerizer = containerizer
        self.container_registry = container_registry
        super().__init__(**kwargs)

    @operation()
    def create_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        """Create the service."""
        raise NotImplementedError

    @operation()
    def delete_service(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete the service."""
        raise NotImplementedError

    @operation()
    def get_service(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        """Get the service."""
        raise NotImplementedError

    @operation()
    def list_services(
        self,
        **kwargs: Any,
    ) -> Response[List[ServiceItem]]:
        """List deployed services."""
        raise NotImplementedError

    @operation()
    def get_logs(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[None]:
        """Get the logs of the service."""
        raise NotImplementedError

    @operation()
    def list_revisions(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[List[Revision]]:
        """List revisions of the service."""
        raise NotImplementedError

    @operation()
    def get_revision(
        self,
        name: str,
        revision: str,
        **kwargs: Any,
    ) -> Response[Revision]:
        """Get a specific revision of the service."""
        raise NotImplementedError

    @operation()
    def delete_revision(
        self,
        name: str,
        revision: str,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete a specific revision of the service."""
        raise NotImplementedError

    @operation()
    def update_traffic(
        self,
        name: str,
        traffic: list[TrafficAllocation],
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        """Update traffic distribution for the service."""
        raise NotImplementedError

    @operation()
    def close(self) -> None:
        """Close the component."""
        raise NotImplementedError

    @operation()
    async def acreate_service(
        self,
        service: ServiceDefinition | None = None,
        where: str | Expression | None = None,
        run_context: RunContext = RunContext(),
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        """Create the service."""
        raise NotImplementedError

    @operation()
    async def adelete_service(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete the service."""
        raise NotImplementedError

    @operation()
    async def aget_service(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        """Get the service."""
        raise NotImplementedError

    @operation()
    async def alist_services(
        self,
        **kwargs: Any,
    ) -> Response[List[ServiceItem]]:
        """List deployed services."""
        raise NotImplementedError

    @operation()
    async def aget_logs(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[None]:
        """Get the logs of the service."""
        raise NotImplementedError

    @operation()
    async def alist_revisions(
        self,
        name: str,
        **kwargs: Any,
    ) -> Response[List[Revision]]:
        """List revisions of the service."""
        raise NotImplementedError

    @operation()
    async def aget_revision(
        self,
        name: str,
        revision: str,
        **kwargs: Any,
    ) -> Response[Revision]:
        """Get a specific revision of the service."""
        raise NotImplementedError

    @operation()
    async def adelete_revision(
        self,
        name: str,
        revision: str,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete a specific revision of the service."""
        raise NotImplementedError

    @operation()
    async def aupdate_traffic(
        self,
        name: str,
        traffic: list[TrafficAllocation],
        **kwargs: Any,
    ) -> Response[ServiceItem]:
        """Update traffic distribution for the service."""
        raise NotImplementedError

    @operation()
    async def aclose(self) -> None:
        """Close the component."""
        raise NotImplementedError



"""
Document Store on MongoDB.
"""

from __future__ import annotations

__all__ = ["MongoDB"]

import copy
import re
from typing import Any

from pymongo import DeleteOne, IndexModel, ReplaceOne
from pymongo.errors import (
    CollectionInvalid,
    DuplicateKeyError,
    OperationFailure,
)

from x8.core import Context, NCall, Operation, Response
from x8.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    OrderByDirection,
    QueryFunctionName,
    Select,
    Update,
    UpdateOp,
    Value,
)
from x8.storage._common import (
    ArrayIndex,
    AscIndex,
    CollectionResult,
    CollectionStatus,
    CompositeIndex,
    DescIndex,
    FieldIndex,
    GeospatialIndex,
    HashIndex,
    Index,
)
from x8.storage._common import IndexHelper as BaseIndexHelper
from x8.storage._common import (
    IndexResult,
    IndexStatus,
    ItemProcessor,
    ParameterParser,
    RangeIndex,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    TextIndex,
    TTLIndex,
    Validator,
    WildcardIndex,
)

from .._helper import (
    build_item_from_parts,
    build_item_from_value,
    build_query_result,
    get_collection_config,
)
from .._models import DocumentCollectionConfig, DocumentFieldType


class MongoDB(StoreProvider):
    uri: str
    database: str
    collection: str | None
    id_map_field: str | dict | None
    pk_map_field: str | dict | None
    etag_embed_field: str | dict | None
    suppress_fields: list[str] | None
    nparams: dict[str, Any]

    _mongo_client: Any
    _amongo_client: Any
    _database_client: Any
    _adatabase_client: Any
    _collection_cache: dict[str, MongoDBCollection]
    _acollection_cache: dict[str, MongoDBCollection]

    def __init__(
        self,
        uri: str,
        database: str,
        collection: str | None = None,
        id_map_field: str | dict | None = "id",
        pk_map_field: str | dict | None = "pk",
        etag_embed_field: str | dict | None = "_etag",
        suppress_fields: list[str] | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            uri:
                URI to MongoDB instance which embeds the credentials.
            database:
                MongoDB database name.
            collection:
                MongoDB collection name mapped to document store collection.
            id_map_field:
                Field in the document to map into id.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            pk_map_field:
                Field in the document to map into pk.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            etag_embed_field:
                Field to store the generated ETAG value.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field.
            suppress_fields:
                List of fields to supress when results are returned.
            nparams:
                Native parameters to pymongo client.
        """
        self.uri = uri
        self.database = database
        self.collection = collection
        self.id_map_field = id_map_field
        self.pk_map_field = pk_map_field
        self.etag_embed_field = etag_embed_field
        self.suppress_fields = suppress_fields
        self.nparams = nparams

        self._mongo_client = None
        self._amongo_client = None
        self._database_client = None
        self._adatabase_client = None
        self._collection_cache = dict()
        self._acollection_cache = dict()

    def __setup__(self, context: Context | None = None) -> None:
        if self._mongo_client is not None:
            return

        from pymongo import MongoClient
        from pymongo.server_api import ServerApi

        self._mongo_client = MongoClient(
            self.uri,
            server_api=ServerApi("1"),
            **self.nparams,
        )
        self._database_client = self._mongo_client[self.database]

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._amongo_client is not None:
            return

        import motor.motor_asyncio

        self._amongo_client = motor.motor_asyncio.AsyncIOMotorClient(
            self.uri,
            **self.nparams,
        )
        self._adatabase_client = self._amongo_client[self.database]

    def _get_collection_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        db_collection = (
            collection_name or self.collection or self.__component__.collection
        )
        return db_collection

    def _get_collections(
        self, op_parser: StoreOperationParser
    ) -> list[MongoDBCollection]:
        if op_parser.is_resource_op():
            return []
        if op_parser.op_equals(StoreOperation.TRANSACT):
            collections: list[MongoDBCollection] = []
            for single_op_parser in op_parser.get_operation_parsers():
                collections.extend(self._get_collections(single_op_parser))
            return collections
        db_collection = self._get_collection_name(op_parser)
        if db_collection is None:
            raise BadRequestError("Collection name must be specified")
        if db_collection in self._collection_cache:
            return [self._collection_cache[db_collection]]
        self._database_client = self._mongo_client[self.database]
        client = self._database_client[db_collection]
        id_map_field = ParameterParser.get_collection_parameter(
            self.id_map_field or self.__component__.id_map_field,
            db_collection,
        )
        pk_map_field = ParameterParser.get_collection_parameter(
            self.pk_map_field or self.__component__.pk_map_field,
            db_collection,
        )
        etag_embed_field = ParameterParser.get_collection_parameter(
            self.etag_embed_field, db_collection
        )
        col = MongoDBCollection(
            client,
            ClientHelper,
            id_map_field,
            pk_map_field,
            etag_embed_field,
            self.suppress_fields,
        )
        self._collection_cache[db_collection] = col
        return [col]

    async def _aget_collections(
        self, op_parser: StoreOperationParser
    ) -> list[MongoDBCollection]:
        if op_parser.is_resource_op():
            return []
        if op_parser.op_equals(StoreOperation.TRANSACT):
            collections: list[MongoDBCollection] = []
            for single_op_parser in op_parser.get_operation_parsers():
                collections.extend(
                    await self._aget_collections(single_op_parser)
                )
            return collections
        db_collection = self._get_collection_name(op_parser)
        if db_collection is None:
            raise BadRequestError("Collection name must be specified")
        if db_collection in self._acollection_cache:
            return [self._acollection_cache[db_collection]]
        self._adatabase_client = self._amongo_client[self.database]
        client = self._adatabase_client[db_collection]
        id_map_field = ParameterParser.get_collection_parameter(
            self.id_map_field or self.__component__.id_map_field,
            db_collection,
        )
        pk_map_field = ParameterParser.get_collection_parameter(
            self.pk_map_field or self.__component__.pk_map_field,
            db_collection,
        )
        etag_embed_field = ParameterParser.get_collection_parameter(
            self.etag_embed_field, db_collection
        )
        col = MongoDBCollection(
            client,
            AsyncClientHelper,
            id_map_field,
            pk_map_field,
            etag_embed_field,
            self.suppress_fields,
        )
        self._acollection_cache[db_collection] = col
        return [col]

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[StoreOperation.PUT, StoreOperation.DELETE],
                single_collection=True,
            )

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collections = self._get_collections(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collections,
            ResourceHelper(self._mongo_client, self._database_client),
        )
        if ncall is None:
            return super().__run__(
                operation,
                context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
            collections,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.__asetup__(context=context)
        op_parser = StoreOperationParser(operation)
        self._validate(op_parser)
        collections = await self._aget_collections(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collections,
            AsyncResourceHelper(self._amongo_client, self._adatabase_client),
        )
        if ncall is None:
            return await super().__arun__(
                operation,
                context,
                **kwargs,
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
            collections,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collections: list[MongoDBCollection],
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if len(collections) == 1:
            op_converter = collections[0].op_converter
            client = collections[0].client
            helper = collections[0].helper
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "config": get_collection_config(op_parser),
                "exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # HAS COLLECTION
        elif op_parser.op_equals(StoreOperation.HAS_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.has_collection, args)
        # CREATE INDEX
        elif op_parser.op_equals(StoreOperation.CREATE_INDEX):
            args = {
                "index": op_parser.get_index(),
                "exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(helper.create_index, args, None)
        # DROP INDEX
        elif op_parser.op_equals(StoreOperation.DROP_INDEX):
            args = {
                "index": op_parser.get_index(),
                "exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(helper.drop_index, args, None)
        # LIST INDEXES
        elif op_parser.op_equals(StoreOperation.LIST_INDEXES):
            args = {"nargs": nargs}
            call = NCall(helper.list_indexes, args, None)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = op_converter.convert_get(op_parser.get_key())
            call = NCall(client.find_one, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args, state, func = op_converter.convert_put(
                op_parser.get_key(),
                op_parser.get_value(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
            )
            if func == "replace_one":
                call = NCall(client.replace_one, args, nargs)
            elif func == "insert_one":
                call = NCall(
                    client.insert_one,
                    args,
                    nargs,
                    {DuplicateKeyError: PreconditionFailedError},
                )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            (args, state, func) = op_converter.convert_update(
                op_parser.get_key(),
                op_parser.get_set(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
                op_parser.get_returning_as_bool(),
            )
            if func == "update_one":
                call = NCall(client.update_one, args, nargs)
            elif func == "find_one_and_update":
                call = NCall(client.find_one_and_update, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = op_converter.convert_delete(
                op_parser.get_key(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
            )
            call = NCall(client.delete_one, args, nargs)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = op_converter.convert_query(
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            args = {"args": args, "nargs": nargs}
            call = NCall(helper.query, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = op_converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(client.count_documents, args, nargs)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, state = op_converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            call = NCall(helper.batch, args, None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            args, state = OperationConverter.convert_transact(
                op_parser.get_operation_parsers(),
                [col.op_converter for col in collections],
            )
            call = NCall(resource_helper.transact, args, None)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collections: list[MongoDBCollection],
    ) -> Any:
        if len(collections) == 1:
            processor = collections[0].processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = nresult
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = nresult
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = nresult
        # HAS COLLECTION
        elif op_parser.op_equals(StoreOperation.HAS_COLLECTION):
            result = nresult
        # CREATE INDEX
        if op_parser.op_equals(StoreOperation.CREATE_INDEX):
            result = nresult
        # DROP INDEX
        elif op_parser.op_equals(StoreOperation.DROP_INDEX):
            result = nresult
        # LIST INDEXES
        elif op_parser.op_equals(StoreOperation.LIST_INDEXES):
            result = nresult
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            if nresult is None:
                raise NotFoundError
            result = build_item_from_value(
                processor=processor,
                value=nresult,
                include_value=True,
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            where = op_parser.get_where()
            exists = op_parser.get_where_exists()
            returning = op_parser.get_returning_as_bool()
            if exists is False:
                pass
            elif where is not None:
                if nresult.matched_count == 0:
                    raise PreconditionFailedError
            value: Any = state["value"] if state is not None else None
            result = build_item_from_value(
                processor=processor,
                value=value,
                include_value=(returning or False),
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            where = op_parser.get_where()
            returning = op_parser.get_returning_as_bool()
            if where is None:
                if returning is None:
                    if nresult.matched_count == 0:
                        raise NotFoundError
                else:
                    if nresult is None:
                        raise NotFoundError
            elif where is not None:
                if returning is None:
                    if nresult.matched_count == 0:
                        raise PreconditionFailedError
                else:
                    if nresult is None:
                        raise PreconditionFailedError
            if isinstance(nresult, dict):
                result = build_item_from_value(
                    processor=processor, value=nresult, include_value=True
                )
            else:
                key = op_parser.get_key()
                result = build_item_from_parts(
                    processor=processor,
                    key=key,
                    etag=(
                        state["etag"]
                        if state is not None and "etag" in state
                        else None
                    ),
                )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            where = op_parser.get_where()
            if where is None:
                if nresult.deleted_count == 0:
                    raise NotFoundError
            elif where is not None:
                if nresult.deleted_count == 0:
                    raise PreconditionFailedError
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items: list = []
            for item in nresult:
                items.append(
                    build_item_from_value(
                        processor=processor, value=item, include_value=True
                    )
                )
            result = build_query_result(items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = nresult
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = []
            if state is not None:
                for st in state["values"]:
                    if st is not None:
                        result.append(
                            build_item_from_value(
                                processor=processor,
                                value=st["value"],
                            )
                        )
                    else:
                        result.append(None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            result = []
            op_parsers = op_parser.get_operation_parsers()
            for i in range(0, len(op_parsers)):
                if state is not None:
                    st = state["values"][i]
                result.append(
                    self._convert_nresult(
                        nresult[i], st, op_parsers[i], [collections[i]]
                    )
                )
        return result


class ResultConverter:
    processor: ItemProcessor

    def __init__(self, processor: ItemProcessor):
        self.processor = processor

    @staticmethod
    def _convert_index(mindex: Any, use_name_type: bool) -> Index:
        name = mindex["name"]
        name_type = BaseIndexHelper.get_type_from_name(name)
        key = mindex["key"]
        if len(key) == 1:
            field, order = next(iter(key.items()))
            if field.endswith("$**"):
                excluded = []
                if "wildcardProjection" in mindex:
                    for ex_field, ex_order in mindex[
                        "wildcardProjection"
                    ].items():
                        if ex_order == 0:
                            excluded.append(ex_field)
                return WildcardIndex(
                    field=field.replace("$**", "*"),
                    excluded=excluded,
                    name=name,
                )
            elif order == 1:
                if "expireAfterSeconds" in mindex:
                    return TTLIndex(field=field, name=name)
                elif use_name_type and name_type == "array":
                    return ArrayIndex(field=field, name=name)
                elif use_name_type and name_type == "range":
                    return RangeIndex(field=field, name=name)
                elif use_name_type and name_type == "field":
                    return FieldIndex(field=field, name=name)
                else:
                    return AscIndex(field=field, name=name)
            elif order == -1:
                return DescIndex(field=field, name=name)
            elif order == "hashed":
                return HashIndex(field=field, name=name)
            elif order == "2dsphere":
                return GeospatialIndex(field=field, name=name)
            return FieldIndex(field=field, name=name)
        else:
            if "weights" in mindex:
                field, order = field, order = next(
                    iter(mindex["weights"].items())
                )
                nconfig = dict()
                if "default_language" in mindex:
                    nconfig["default_language"] = mindex["default_language"]
                if "language_override" in mindex:
                    nconfig["language_override"] = mindex["language_override"]
                nconfig["weights"] = mindex["weights"]
                return TextIndex(field=field, nconfig=nconfig, name=name)
            else:
                field_indexes: list = []
                name_types: list = (
                    BaseIndexHelper.get_composite_types_from_name(name)
                )
                if len(key.items()) != len(name_types):
                    name_types = [None] * len(key.items())
                for (field, order), name_type in zip(key.items(), name_types):
                    if order == 1:
                        if use_name_type and name_type == "array":
                            field_indexes.append(ArrayIndex(field=field))
                        elif use_name_type and name_type == "hash":
                            field_indexes.append(HashIndex(field=field))
                        elif use_name_type and name_type == "range":
                            field_indexes.append(RangeIndex(field=field))
                        elif use_name_type and name_type == "field":
                            field_indexes.append(FieldIndex(field=field))
                        else:
                            field_indexes.append(AscIndex(field=field))
                    elif order == -1:
                        field_indexes.append(DescIndex(field=field))
                    elif order == "2dsphere":
                        field_indexes.append(GeospatialIndex(field=field))
                return CompositeIndex(fields=field_indexes, name=name)

    @staticmethod
    def convert_list_indexes(nresult: Any, use_name_type: bool) -> list[Index]:
        indexes: list = []
        for mindex in nresult:
            if mindex["name"] == "_id_":
                continue
            indexes.append(
                ResultConverter._convert_index(mindex, use_name_type)
            )
        return indexes


class OperationConverter:
    processor: ItemProcessor
    client: Any

    def __init__(self, processor: ItemProcessor, client: Any):
        self.processor = processor
        self.client = client

    @staticmethod
    def convert_create_indexes(indexes: list[Index]) -> tuple[dict, list]:
        index_results = []
        mindexes = []
        for index in indexes:
            args = OperationConverter.convert_create_index(index)
            if args is not None:
                index_results.append(IndexResult(status=IndexStatus.CREATED))
                mindexes.append(IndexModel(**args))
            else:
                index_results.append(
                    IndexResult(status=IndexStatus.NOT_SUPPORTED)
                )
        return {"indexes": mindexes}, index_results

    @staticmethod
    def convert_create_index(index: Index) -> dict | None:
        args: dict = {"name": BaseIndexHelper.convert_index_name(index)}
        keys: Any = None
        if isinstance(index, (FieldIndex, RangeIndex, ArrayIndex)):
            field = OperationConverter._convert_field(index.field)
            keys = field
        elif isinstance(index, AscIndex):
            field = OperationConverter._convert_field(index.field)
            keys = [(field, 1)]
        elif isinstance(index, DescIndex):
            field = OperationConverter._convert_field(index.field)
            keys = [(field, -1)]
        elif isinstance(index, HashIndex):
            field = OperationConverter._convert_field(index.field)
            keys = [(field, "hashed")]
        elif isinstance(index, CompositeIndex):
            keys = []
            for field_index in index.fields:
                field = OperationConverter._convert_field(field_index.field)
                if isinstance(field_index, DescIndex):
                    keys.append((field, -1))
                elif isinstance(field_index, TextIndex):
                    keys.append((field, "text"))
                elif isinstance(field_index, GeospatialIndex):
                    keys.append((field, "2dsphere"))
                else:
                    keys.append((field, 1))
        elif isinstance(index, WildcardIndex):
            field = OperationConverter._convert_field(index.field)
            if field.endswith("*"):
                field = field.replace("*", "$**")
            else:
                field = f"{field}.$**"
            keys = [(field, 1)]
            projection: dict = {}
            for exclude in index.excluded:
                projection[exclude] = 0
            args["wildcardProjection"] = projection
        elif isinstance(index, TextIndex):
            field = OperationConverter._convert_field(index.field)
            keys = [(field, "text")]
        elif isinstance(index, GeospatialIndex):
            field = OperationConverter._convert_field(index.field)
            keys = [(field, "2dsphere")]
        elif isinstance(index, TTLIndex):
            field = OperationConverter._convert_field(index.field)
            keys = [(field, 1)]
            args["expireAfterSeconds"] = 0
        else:
            return None
        args["keys"] = keys
        if index.nconfig:
            args = args | index.nconfig
        return args

    @staticmethod
    def convert_drop_index(index: Index) -> dict:
        return {"index_or_name": BaseIndexHelper.convert_index_name(index)}

    @staticmethod
    def convert_transact(
        op_parsers: list[StoreOperationParser],
        converters: list[OperationConverter],
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for i in range(0, len(op_parsers)):
            op_parser = op_parsers[i]
            converter = converters[i]
            if op_parser.op_equals(StoreOperation.PUT):
                args, state, func = converter.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(
                    {"func": func, "args": args, "client": converter.client}
                )
                states.append(state)
            elif op_parser.op_equals(StoreOperation.UPDATE):
                (args, state, func) = converter.convert_update(
                    op_parser.get_key(),
                    op_parser.get_set(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                    op_parser.get_returning_as_bool(),
                )
                ops.append(
                    {"func": func, "args": args, "client": converter.client}
                )
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args = converter.convert_delete(
                    op_parser.get_key(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(
                    {
                        "func": "delete_one",
                        "args": args,
                        "client": converter.client,
                    }
                )
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.PUT):
                args, state, func = self.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(ReplaceOne(**args))
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args = self.convert_delete(
                    op_parser.get_key(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(DeleteOne(**args))
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_get(self, key: Value) -> dict:
        filter = self.processor.get_key_from_key(key)
        args = {"filter": filter}
        return args

    def convert_put(
        self,
        key: Value,
        value: dict,
        where: Expression | None,
        exists: bool | None,
    ) -> tuple[dict, dict | None, str]:
        document = self.processor.add_embed_fields(value, key)
        if key is not None:
            filter = self.processor.get_key_from_key(key)
        else:
            filter = self.processor.get_key_from_value(document)
        if where is None:
            args = {
                "filter": filter,
                "replacement": document,
                "upsert": True,
            }
            func = "replace_one"
        elif exists is False:
            args = {"document": document}
            func = "insert_one"
        elif exists is True:
            args = {
                "filter": filter,
                "replacement": document,
            }
            func = "replace_one"
        elif where is not None:
            where_filter = self.convert_expr(where)
            and_filter = {"$and": [filter, where_filter]}
            args = {
                "filter": and_filter,
                "replacement": document,
            }
            func = "replace_one"
        return args, {"value": document}, func

    def convert_update(
        self,
        key: Value,
        set: Update,
        where: Expression | None,
        exists: bool | None,
        returning: bool | None,
    ) -> tuple[dict, dict | None, str]:
        uset = copy.deepcopy(set)
        state = None
        filter = self.processor.get_key_from_key(key)
        if self.processor.needs_local_etag():
            etag = self.processor.generate_etag()
            uset = self.processor.add_etag_update(uset, etag)
            state = {"etag": etag}
        updates = self.convert_update_ops(uset)
        if where is None or exists is True:
            args: dict = {"filter": filter, "update": updates}
        elif where is not None:
            where_filter = self.convert_expr(where)
            and_filter = {"$and": [filter, where_filter]}
            args = {"filter": and_filter, "update": updates}
        if returning is None:
            func = "update_one"
        else:
            args = args | {"return_document": returning}
            func = "find_one_and_update"
        return args, state, func

    def convert_delete(
        self, key: Value, where: Expression | None, exists: bool | None
    ) -> dict:
        filter = self.processor.get_key_from_key(key)
        if where is None or exists is True:
            args = {"filter": filter}
        elif where is not None:
            where_filter = self.convert_expr(where)
            and_filter = {"$and": [filter, where_filter]}
            args = {"filter": and_filter}
        return args

    def convert_query(
        self,
        select: Select | None = None,
        where: Expression | None = None,
        order_by: OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        args = {}
        if select is not None:
            args["projection"] = self.convert_select(select)
        if where is not None:
            args["filter"] = self.convert_expr(where)
        if order_by is not None:
            args["sort"] = self.convert_order_by(order_by)
        if limit is not None:
            args["limit"] = limit
        if offset is not None:
            args["skip"] = offset
        return args

    def convert_count(self, where: Expression | None = None) -> dict:
        args = {}
        if where is not None:
            args["filter"] = self.convert_expr(where)
        else:
            args["filter"] = {}
        return args

    @staticmethod
    def _convert_field(field: str) -> str:
        return field.replace("[", ".").replace("]", "").rstrip(".")

    def convert_field(self, field: Field | str) -> str:
        path: str
        if isinstance(field, str):
            path = self.processor.resolve_field(field)
        elif isinstance(field, Field):
            path = self.processor.resolve_field(field.path)
        return OperationConverter._convert_field(path)

    def convert_comparison(self, expr: Comparison) -> dict:
        func = False
        if isinstance(expr.lexpr, Field):
            lhs = self.convert_field(expr.lexpr)
            rhs = expr.rexpr
        elif isinstance(expr.rexpr, Field):
            lhs = self.convert_field(expr.rexpr)
            rhs = expr.lexpr
        elif isinstance(expr.lexpr, Function):
            func = True
            lhs = self.convert_func(expr.lexpr)
            rhs = expr.rexpr
        elif isinstance(expr.rexpr, Function):
            func = True
            lhs = self.convert_func(expr.rexpr)
            rhs = expr.lexpr
        else:
            raise BadRequestError("Comparison not supported")
        ops_map = {
            ComparisonOp.LT: "$lt",
            ComparisonOp.LTE: "$lte",
            ComparisonOp.GT: "$gt",
            ComparisonOp.GTE: "$gte",
            ComparisonOp.EQ: "$eq",
            ComparisonOp.NEQ: "$ne",
            ComparisonOp.IN: "$in",
            ComparisonOp.NIN: "$nin",
        }
        op = expr.op
        if op in ops_map:
            nop = ops_map[op]
            if not func:
                return {lhs: {nop: rhs}}
            else:
                return {"$expr": {nop: [lhs, rhs]}}
        else:
            if op is ComparisonOp.BETWEEN:
                if isinstance(rhs, list):
                    return {lhs: {"$gte": rhs[0], "$lte": rhs[1]}}
            elif op is ComparisonOp.LIKE:
                regex = re.escape(str(rhs))
                return {lhs: {"$regex": regex}}
        raise BadRequestError(f"Comparison {expr} not supported")

    def convert_func(self, expr: Function) -> Any:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == QueryFunctionName.IS_TYPE:
                field = self.convert_field(args[0])
                type = args[1]
                if isinstance(type, str):
                    type = DocumentFieldType(type.lower())
                if type == DocumentFieldType.STRING:
                    return {field: {"$type": "string"}}
                elif type == DocumentFieldType.NUMBER:
                    return {
                        "$or": [
                            {field: {"$type": "double"}},
                            {field: {"$type": "int"}},
                            {field: {"$type": "long"}},
                            {field: {"$type": "decimal"}},
                        ]
                    }
                elif type == DocumentFieldType.BOOLEAN:
                    return {field: {"$type": "bool"}}
                elif type == DocumentFieldType.OBJECT:
                    return {field: {"$type": "object"}}
                elif type == DocumentFieldType.ARRAY:
                    return {field: {"$type": "array"}}
                elif type == DocumentFieldType.NULL:
                    return {field: {"$type": "null"}}
            if name == QueryFunctionName.IS_DEFINED:
                field = self.convert_field(args[0])
                return {field: {"$exists": True}}
            if name == QueryFunctionName.IS_NOT_DEFINED:
                field = self.convert_field(args[0])
                return {field: {"$exists": False}}
            if name == QueryFunctionName.LENGTH:
                field = self.convert_field(args[0])
                splits = field.split(".")
                if splits[-1].isnumeric():
                    arr_field = str.join(".", splits[0:-1])
                    arr_index = int(splits[-1])
                    return {
                        "$strLenCP": {
                            "$arrayElemAt": [f"${arr_field}", arr_index]
                        }
                    }
                else:
                    return {"$strLenCP": f"${field}"}
            if name == QueryFunctionName.CONTAINS:
                field = self.convert_field(args[0])
                regex = re.escape(args[1])
                return {field: {"$regex": regex}}
            if name == QueryFunctionName.STARTS_WITH:
                field = self.convert_field(args[0])
                regex = re.escape(args[1])
                return {field: {"$regex": f"^{regex}"}}
            if name == QueryFunctionName.ARRAY_LENGTH:
                field = self.convert_field(args[0])
                return {"$size": f"${field}"}
            if name == QueryFunctionName.ARRAY_CONTAINS:
                field = self.convert_field(args[0])
                return {field: {"$all": [args[1]]}}
            if name == QueryFunctionName.ARRAY_CONTAINS_ANY:
                field = self.convert_field(args[0])
                return {field: {"$in": args[1]}}
        raise BadRequestError(f"Function {name} not recognized")

    def convert_expr(self, expr: Expression | None) -> Any:
        if expr is None or isinstance(
            expr, (str, int, float, bool, dict, list)
        ):
            return expr
        if isinstance(expr, Field):
            return self.convert_field(expr)
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            return self.convert_comparison(expr)
        if isinstance(expr, And):
            return {
                "$and": [
                    self.convert_expr(expr.lexpr),
                    self.convert_expr(expr.rexpr),
                ]
            }
        if isinstance(expr, Or):
            return {
                "$or": [
                    self.convert_expr(expr.lexpr),
                    self.convert_expr(expr.rexpr),
                ]
            }
        if isinstance(expr, Not):
            return {"$nor": [self.convert_expr(expr.expr)]}
        raise BadRequestError(f"Expression {expr!r} not supported")

    def convert_order_by(self, order_by: OrderBy) -> Any:
        sort_list = []
        for term in order_by.terms:
            field = self.convert_field(term.field)
            if (
                term.direction is None
                or term.direction == OrderByDirection.ASC
            ):
                sort_list.append((field, 1))
            elif term.direction == OrderByDirection.DESC:
                sort_list.append((field, -1))
        return sort_list

    def convert_select(self, select: Select) -> Any:
        if len(select.terms) == 0:
            return None
        field_paths = []
        for term in select.terms:
            field_path = self.convert_field(term.field)
            field_paths.append(field_path)
        return field_paths

    def convert_update_ops(self, update: Update) -> dict:
        set = {}
        unset = {}
        inc = {}
        rename = {}
        push = {}
        pop = {}
        pull = {}
        add_to_set = {}

        for operation in update.operations:
            field = self.convert_field(Field(path=operation.field))
            splits = field.split(".")
            op = operation.op
            if splits[-1].isnumeric():
                array_field = str.join(".", splits[:-1])
                array_index = int(splits[-1])
                if op == UpdateOp.PUT:
                    set[field] = operation.args[0]
                elif op == UpdateOp.INSERT:
                    if isinstance(operation.args[0], list):
                        arr = operation.args[0]
                    else:
                        arr = [operation.args[0]]
                    push[array_field] = {
                        "$each": arr,
                        "$position": array_index,
                    }
                elif op == UpdateOp.INCREMENT:
                    inc[field] = operation.args[0]
                elif op == UpdateOp.DELETE:
                    if array_index == 0:
                        pop[array_field] = -1
                    else:
                        set[array_field] = {
                            "$concatArrays": [
                                {"$slice": [f"${array_field}", array_index]},
                                {
                                    "$slice": [
                                        f"${array_field}",
                                        array_index + 1,
                                        {"$size": f"${array_field}"},
                                    ]
                                },
                            ]
                        }
            elif splits[-1] == "-":
                array_field = str.join(".", splits[:-1])
                if op == UpdateOp.INSERT:
                    if isinstance(operation.args[0], list):
                        arr = operation.args[0]
                    else:
                        arr = [operation.args[0]]
                    push[array_field] = {
                        "$each": arr,
                    }
                elif op == UpdateOp.DELETE:
                    pop[array_field] = 1
            else:
                if op == UpdateOp.PUT:
                    set[field] = operation.args[0]
                elif op == UpdateOp.INSERT:
                    set[field] = operation.args[0]
                elif op == UpdateOp.INCREMENT:
                    inc[field] = operation.args[0]
                elif op == UpdateOp.DELETE:
                    unset[field] = ""
                elif op == UpdateOp.MOVE:
                    rename[field] = operation.args[0].path
                elif op == UpdateOp.ARRAY_UNION:
                    add_to_set[field] = {"$each": operation.args[0]}
                elif op == UpdateOp.ARRAY_REMOVE:
                    pull[field] = {"$in": operation.args[0]}

        res = {}
        if len(set) > 0:
            res["$set"] = set
        if len(unset) > 0:
            res["$unset"] = unset
        if len(inc) > 0:
            res["$inc"] = inc
        if len(rename) > 0:
            res["$rename"] = rename
        if len(push) > 0:
            res["$push"] = push
        if len(pull) > 0:
            res["$pull"] = pull
        if len(pop) > 0:
            res["$pop"] = pop
        if len(add_to_set) > 0:
            res["$addToSet"] = add_to_set
        return res


class ResourceHelper:
    mongo_client: Any
    database_client: Any

    def __init__(self, mongo_client: Any, database_client: Any):
        self.mongo_client = mongo_client
        self.database_client = database_client

    def create_collection(
        self,
        collection: str,
        config: DocumentCollectionConfig | None,
        exists: bool | None,
        nargs: Any,
    ) -> CollectionResult:
        status = CollectionStatus.CREATED
        try:
            NCall(
                self.database_client.create_collection,
                {"name": collection},
                nargs,
            ).invoke()
        except CollectionInvalid:
            status = CollectionStatus.EXISTS
            if exists is False:
                raise ConflictError

        index_results: list = []
        if config and config.indexes:
            args, index_results = OperationConverter.convert_create_indexes(
                config.indexes
            )
            client = self.database_client[collection]
            NCall(
                client.create_indexes,
                args,
                nargs,
            ).invoke()
        return CollectionResult(status=status, indexes=index_results)

    def drop_collection(
        self,
        collection: str,
        exists: bool | None,
        nargs: Any,
    ) -> CollectionResult:
        response = NCall(
            self.database_client.drop_collection,
            {"name_or_collection": collection},
            nargs,
        ).invoke()
        if "ns" not in response:
            if exists is True:
                raise NotFoundError
            return CollectionResult(status=CollectionStatus.NOT_EXISTS)
        return CollectionResult(status=CollectionStatus.DROPPED)

    def list_collections(self, nargs) -> Any:
        nresult = NCall(
            self.database_client.list_collection_names, None, nargs
        ).invoke()
        return nresult

    def has_collection(
        self,
        collection: str,
        nargs: Any,
    ):
        nresult = NCall(
            self.database_client.list_collection_names, None, nargs
        ).invoke()
        return collection in nresult

    def transact(self, ops: list[dict]) -> Any:
        nresult = []
        with self.mongo_client.start_session() as session:
            with session.start_transaction():
                for op in ops:
                    func = op["func"]
                    client = op["client"]
                    args = op["args"]
                    args["session"] = session
                    if func == "insert_one":
                        res = client.insert_one(**args)
                    elif func == "replace_one":
                        res = client.replace_one(**args)
                        if "upsert" not in args and res.matched_count == 0:
                            raise ConflictError
                    elif func == "update_one":
                        res = client.update_one(**args)
                        if res.matched_count == 0:
                            raise ConflictError
                    elif func == "find_one_and_update":
                        res = client.find_one_and_update(**args)
                        if res is None:
                            raise ConflictError
                    elif func == "delete_one":
                        res = client.delete_one(**args)
                        if res.deleted_count == 0:
                            raise ConflictError
                    nresult.append(res)
                return nresult

    def close(self, nargs: Any) -> Any:
        pass


class AsyncResourceHelper:
    mongo_client: Any
    database_client: Any

    def __init__(self, mongo_client: Any, database_client: Any):
        self.mongo_client = mongo_client
        self.database_client = database_client

    async def create_collection(
        self,
        collection: str,
        config: DocumentCollectionConfig | None,
        exists: bool | None,
        nargs: Any,
    ) -> CollectionResult:
        status = CollectionStatus.CREATED
        try:
            await NCall(
                self.database_client.create_collection,
                {"name": collection},
                nargs,
            ).ainvoke()
        except CollectionInvalid:
            status = CollectionStatus.EXISTS
            if exists is False:
                raise ConflictError

        index_results: list = []
        if config and config.indexes:
            args, index_results = OperationConverter.convert_create_indexes(
                config.indexes
            )
            client = self.database_client[collection]
            await NCall(
                client.create_indexes,
                args,
                nargs,
            ).ainvoke()
        return CollectionResult(status=status, indexes=index_results)

    async def drop_collection(
        self,
        collection: str,
        exists: bool | None,
        nargs: Any,
    ) -> CollectionResult:
        response = await NCall(
            self.database_client.drop_collection,
            {"name_or_collection": collection},
            nargs,
        ).ainvoke()
        if "ns" not in response:
            if exists is True:
                raise NotFoundError
            return CollectionResult(status=CollectionStatus.NOT_EXISTS)
        return CollectionResult(status=CollectionStatus.DROPPED)

    async def list_collections(self, nargs) -> Any:
        nresult = await NCall(
            self.database_client.list_collection_names, None, nargs
        ).ainvoke()
        return nresult

    async def has_collection(
        self,
        collection: str,
        nargs: Any,
    ):
        nresult = await NCall(
            self.database_client.list_collection_names, None, nargs
        ).ainvoke()
        return collection in nresult

    async def transact(self, ops: list[dict]) -> Any:
        nresult = []
        async with await self.mongo_client.start_session() as session:
            async with session.start_transaction():
                for op in ops:
                    func = op["func"]
                    client = op["client"]
                    args = op["args"]
                    args["session"] = session
                    if func == "insert_one":
                        res = await client.insert_one(**args)
                    elif func == "replace_one":
                        res = await client.replace_one(**args)
                        if "upsert" not in args and res.matched_count == 0:
                            raise ConflictError
                    elif func == "update_one":
                        res = await client.update_one(**args)
                        if res.matched_count == 0:
                            raise ConflictError
                    elif func == "find_one_and_update":
                        res = await client.find_one_and_update(**args)
                        if res is None:
                            raise ConflictError
                    elif func == "delete_one":
                        res = await client.delete_one(**args)
                        if res.deleted_count == 0:
                            raise ConflictError
                    nresult.append(res)
                return nresult

    async def close(self, nargs: Any) -> Any:
        pass


class ClientHelper:
    client: Any
    processor: ItemProcessor

    def __init__(self, client: Any, processor: ItemProcessor):
        self.client = client
        self.processor = processor

    def create_index(
        self, index: Index, exists: bool | None, nargs: Any
    ) -> IndexResult:
        indexes = self.list_indexes(nargs, False)
        status, match_index = BaseIndexHelper.check_index_status(
            indexes, index
        )
        if match_index:
            if exists is False:
                raise ConflictError
            return IndexResult(status=status, index=match_index)
        args = OperationConverter.convert_create_index(index)
        if args is not None:
            NCall(self.client.create_index, args, nargs).invoke()
            return IndexResult(status=IndexStatus.CREATED)
        else:
            return IndexResult(status=IndexStatus.NOT_SUPPORTED)

    def drop_index(
        self, index: Index, exists: bool | None, nargs: Any
    ) -> IndexResult:
        args = OperationConverter.convert_drop_index(index)
        try:
            NCall(self.client.drop_index, args, nargs).invoke()
        except OperationFailure:
            if exists is True:
                raise NotFoundError
            return IndexResult(status=IndexStatus.NOT_EXISTS)
        return IndexResult(status=IndexStatus.DROPPED)

    def list_indexes(self, nargs, use_name_type: bool = True) -> list:
        nresult = NCall(self.client.list_indexes, None, nargs).invoke()
        nresult = ResultConverter.convert_list_indexes(nresult, use_name_type)
        return nresult

    def query(self, args, nargs) -> Any:
        nresult = NCall(
            self.client.find,
            args,
            nargs,
        ).invoke()
        return nresult

    def batch(self, ops: list) -> Any:
        nresult = self.client.bulk_write(ops)
        return nresult


class AsyncClientHelper:
    client: Any
    processor: ItemProcessor

    def __init__(self, client: Any, processor: ItemProcessor):
        self.client = client
        self.processor = processor

    async def create_index(
        self, index: Index, exists: bool | None, nargs: Any
    ) -> IndexResult:
        indexes = await self.list_indexes(nargs, False)
        status, match_index = BaseIndexHelper.check_index_status(
            indexes, index
        )
        if match_index:
            if exists is False:
                raise ConflictError
            return IndexResult(status=status, index=match_index)
        args = OperationConverter.convert_create_index(index)
        if args is not None:
            await NCall(self.client.create_index, args, nargs).ainvoke()
            return IndexResult(status=IndexStatus.CREATED)
        else:
            return IndexResult(status=IndexStatus.NOT_SUPPORTED)

    async def drop_index(
        self, index: Index, exists: bool | None, nargs: Any
    ) -> IndexResult:
        args = OperationConverter.convert_drop_index(index)
        try:
            await NCall(self.client.drop_index, args, nargs).ainvoke()
        except OperationFailure:
            if exists is True:
                raise NotFoundError
            return IndexResult(status=IndexStatus.NOT_EXISTS)
        return IndexResult(status=IndexStatus.DROPPED)

    async def list_indexes(self, nargs, use_name_type: bool = True) -> list:
        response = NCall(self.client.list_indexes, None, nargs).invoke()
        nresult = []
        async for item in response:
            nresult.append(item)
        nresult = ResultConverter.convert_list_indexes(nresult, use_name_type)
        return nresult

    async def query(self, args, nargs) -> Any:
        response = NCall(self.client.find, args, nargs).invoke()
        nresult = []
        async for item in response:
            nresult.append(item)
        return nresult

    async def batch(self, ops: list) -> Any:
        nresult = await self.client.bulk_write(ops)
        return nresult


class MongoDBCollection:
    client: Any
    op_converter: OperationConverter
    result_converter: ResultConverter
    processor: ItemProcessor
    helper: Any

    def __init__(
        self,
        client: Any,
        helper_type: Any,
        id_map_field: str | None,
        pk_map_field: str | None,
        etag_embed_field: str | None,
        suppress_fields: list[str] | None,
    ) -> None:
        self.client = client
        self.processor = ItemProcessor(
            id_embed_field="_id",
            etag_embed_field=etag_embed_field,
            id_map_field=id_map_field,
            pk_map_field=pk_map_field,
            local_etag=True,
            suppress_fields=suppress_fields,
        )
        self.op_converter = OperationConverter(
            processor=self.processor, client=self.client
        )
        self.result_converter = ResultConverter(processor=self.processor)
        self.helper = helper_type(client, self.processor)



secrets = [
    {
        "id": "test-id-1",
        "values": [
            "test-1-value-1",
            "test-1-value-2",
            "test-1-value-3",
            "test-1-value-4",
        ],
        "metadata": [
            {
                "test-1-tag-1-key-1": "test-1-tag-1-value-1",
                "test-1-tag-1-key-2": "test-1-tag-1-value-2",
                "common-key": "common-value",
            },
            {
                "test-1-tag-2-key-1": "test-1-tag-2-value-1",
                "test-1-tag-2-key-2": "test-1-tag-2-value-2",
            },
        ],
    },
    {
        "id": "test-id-2",
        "values": ["test-2-value-2"],
        "metadata": [
            {
                "test-2-tag-1-key-1": "test-2-tag-1-value-1",
                "test-2-tag-1-key-2": "test-2-tag-1-value-2",
                "common-key": "common-value",
            }
        ],
    },
    {
        "id": "test-id-11",
        "values": ["test-11-value-2"],
        "metadata": [
            {
                "test-11-tag-1-key-1": "test-11-tag-1-value-1",
                "test-11-tag-1-key-2": "test-11-tag-1-value-2",
            }
        ],
    },
]



__all__ = ["APIKey"]

from x8.core import Provider
from x8.core.exceptions import ForbiddenError

from .._models import AuthResult, UserCredential, UserInfo


class APIKey(Provider):
    api_keys: list[str] | str | None

    def __init__(
        self,
        api_keys: list[str] | str | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            api_keys:
                API keys.
        """
        self.api_keys = api_keys
        super().__init__(**kwargs)

    def validate(
        self,
        credential: str | UserCredential,
    ) -> AuthResult:
        if not isinstance(credential, str):
            raise ForbiddenError("API key must be a string")
        if isinstance(self.api_keys, str):
            if credential == self.api_keys:
                return AuthResult(token=credential)
        elif isinstance(self.api_keys, list):
            if credential in self.api_keys:
                return AuthResult(token=credential)
        raise ForbiddenError("Invalid API key")

    def get_user_info(
        self,
        credential: str | UserCredential,
    ) -> UserInfo:
        raise NotImplementedError(
            "API key provider does not support user info"
        )



import os

from x8.core import Component, operation


class SideService(Component):
    @operation()
    def add(self, a: int, b: int) -> int:
        return a + b

    @operation()
    def subtract(self, a: int, b: int) -> int:
        return a - b

    @operation()
    def version(self) -> str:
        return os.getenv("VERSION", "default")



"""
Key Value Store on PostgreSQL.
"""

from __future__ import annotations

__all__ = ["PostgreSQL"]

from datetime import datetime, timezone
from typing import Any

import psycopg
from psycopg import sql

from x8.core import Context, NCall, Operation, Response
from x8.core.exceptions import (
    BadRequestError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import (
    Comparison,
    ComparisonOp,
    Expression,
    Function,
    FunctionNamespace,
    QueryFunctionName,
    Update,
    UpdateOp,
)
from x8.storage._common import (
    Attribute,
    ItemProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    UpdateAttribute,
)

from .._helper import build_item, convert_value, get_collection_name
from .._models import (
    KeyValueItem,
    KeyValueKeyType,
    KeyValueList,
    KeyValueValueType,
)


class PostgreSQL(StoreProvider):
    connection_string: str
    table: str
    collection: str | None
    nparams: dict[str, Any]

    _client: Any
    _aclient: Any
    _processor: ItemProcessor
    _op_converter: OperationConverter
    _result_converter: ResultConverter

    def __init__(
        self,
        connection_string: str,
        table: str = "kv",
        collection: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            connection_string:
                PostgreSQL connection string.
            table:
                Table name that will host the key value store.
            collection:
                Default collection.
            nparams:
                Native parameters to SQLite client.
        """
        self.connection_string = connection_string
        self.table = table
        self.collection = collection
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._processor = ItemProcessor()
        self._op_converter = OperationConverter(self._processor, self.table)
        self._result_converter = ResultConverter(self._processor)

    def __setup__(
        self,
        context: Context | None = None,
    ) -> None:
        if self._client is not None:
            return

        self._client = psycopg.connect(self.connection_string, **self.nparams)
        self._create_table_if_needed()

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return
        self._aclient = await psycopg.AsyncConnection.connect(
            self.connection_string, **self.nparams
        )
        await self._acreate_table_if_needed()

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = self.get_op_parser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            ClientHelper(self._client, self._op_converter),
        )
        if ncall is None:
            return super().__run__(
                operation,
                context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.__asetup__(context=context)
        op_parser = self.get_op_parser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            AsyncClientHelper(self._aclient, self._op_converter),
        )
        if ncall is None:
            return await super().__arun__(
                operation,
                context,
                **kwargs,
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        helper: Any,
    ) -> tuple[NCall | None, dict]:
        call = None
        state: dict = {}
        op_converter = self._op_converter
        collection: str | None = get_collection_name(self, op_parser)
        # EXISTS
        if op_parser.op_equals(StoreOperation.EXISTS):
            args: dict = {
                "key": op_parser.get_key(),
                "collection": collection,
            }
            call = NCall(helper.exists, args)
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            args = {
                "key": op_parser.get_key(),
                "start": op_parser.get_start(),
                "end": op_parser.get_end(),
                "collection": collection,
            }
            call = NCall(helper.get, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args, state = op_converter.convert_put(
                op_parser.get_key(),
                op_parser.get_value(),
                op_parser.get_where_exists(),
                op_parser.get_where_etag(),
                op_parser.get_expiry(),
                op_parser.get_returning(),
                collection,
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {psycopg.errors.UniqueViolation: PreconditionFailedError},
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            args, state = op_converter.convert_update(
                op_parser.get_key(),
                op_parser.get_set(),
                op_parser.get_where_etag(),
                op_parser.get_returning(),
                collection,
            )
            call = NCall(helper.execute, args)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = op_converter.convert_delete(
                op_parser.get_key(),
                op_parser.get_where_etag(),
                collection,
            )
            call = NCall(helper.execute, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = op_converter.convert_query(
                op_parser.get_where(),
                op_parser.get_limit(),
                collection,
            )
            call = NCall(helper.execute, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = op_converter.convert_count(
                op_parser.get_where(),
                collection,
            )
            call = NCall(helper.execute, args)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, state = op_converter.convert_batch(
                op_parser.get_operation_parsers(),
                collection,
            )
            call = NCall(helper.batch, args)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            call = NCall(helper.close, None)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: StoreOperationParser,
    ) -> Any:
        result: Any = None
        result_converter = self._result_converter
        type = self.__component__.type
        # EXISTS
        if op_parser.op_equals(StoreOperation.EXISTS):
            result = nresult
        # GET
        if op_parser.op_equals(StoreOperation.GET):
            result = result_converter.convert_get(nresult, type)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = result_converter.convert_put(
                nresult,
                op_parser.get_key(),
                op_parser.get_value(),
                op_parser.get_where_exists(),
                op_parser.get_where_etag(),
                op_parser.get_returning(),
                state["etag"],
                type,
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            result = result_converter.convert_update(
                nresult,
                op_parser.get_key(),
                op_parser.get_where_etag(),
                op_parser.get_returning(),
                state["etag"],
                type,
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = result_converter.convert_delete(
                nresult,
                op_parser.get_where_etag(),
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            result = result_converter.convert_query(nresult)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = result_converter.convert_count(nresult)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = result_converter.convert_batch(
                nresult,
                op_parser.get_operation_parsers(),
                state["etags"],
                type,
            )
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            pass
        return result

    def _create_table_if_needed(self, **kwargs: Any) -> None:
        cursor = self._client.cursor()
        rows = cursor.execute(
            """SELECT tablename FROM pg_catalog.pg_tables
            WHERE schemaname = 'public' AND tablename = %s""",
            (self.table,),
        ).fetchall()
        if len(rows) == 0:
            cursor.execute(
                sql.SQL(
                    """
                    CREATE TABLE {} (
                    collection TEXT,
                    id TEXT,
                    value BYTEA,
                    etag TEXT,
                    expiry DOUBLE PRECISION,
                    PRIMARY KEY (collection, id)
                    )
                    """
                ).format(sql.Identifier(self.table))
            )
        self._client.commit()
        cursor.close()

    async def _acreate_table_if_needed(self, **kwargs: Any) -> None:
        cursor = self._aclient.cursor()
        await cursor.execute(
            """SELECT tablename FROM pg_catalog.pg_tables
            WHERE schemaname = 'public' AND tablename = %s""",
            (self.table,),
        )
        rows = await cursor.fetchall()
        if len(rows) == 0:
            await cursor.execute(
                sql.SQL(
                    """
                    CREATE TABLE {} (
                    collection TEXT,
                    id TEXT,
                    value BYTEA,
                    etag TEXT,
                    expiry DOUBLE PRECISION,
                    PRIMARY KEY (collection, id)
                    )
                    """
                ).format(sql.Identifier(self.table))
            )
        await self._aclient.commit()
        await cursor.close()


class OperationConverter:
    processor: ItemProcessor
    table: str

    def __init__(self, processor: ItemProcessor, table: str):
        self.processor = processor
        self.table = table

    def convert_exists(
        self,
        key: KeyValueKeyType,
        collection: str | None,
    ) -> dict:
        id = self.processor.get_id_from_key(key)
        query = f"""
                SELECT id, expiry FROM {self.table}
                WHERE collection = %s AND id = %s
                """
        return {
            "query": query,
            "params": (
                collection,
                id,
            ),
        }

    def convert_get(
        self,
        key: KeyValueKeyType,
        start: int | None,
        end: int | None,
        collection: str | None,
    ) -> dict:
        id = self.processor.get_id_from_key(key)
        value_column = "value"
        if start and end:
            value_column = f"substr(value, {start + 1}, {end - start + 1})"
        elif start:
            value_column = f"""substr(
                    value, {start + 1}, LENGTH(value) - {start}
                )"""
        elif end:
            value_column = f"substr(value, 1, {end + 1})"
        query = f"""
                SELECT id, {value_column}, etag, expiry
                FROM {self.table}
                WHERE collection = %s AND id = %s
                """
        return {
            "query": query,
            "params": (
                collection,
                id,
            ),
        }

    def convert_put_value(self, value: KeyValueValueType) -> bytes:
        if isinstance(value, (bytes, memoryview)):
            return value
        return str(value).encode("utf-8")

    def convert_put(
        self,
        key: KeyValueKeyType,
        value: KeyValueValueType,
        where_exists: bool | None,
        where_etag: bool | None,
        expiry: int | None,
        returning: str | None,
        collection: str | None,
    ) -> tuple[dict, dict]:
        id = self.processor.get_id_from_key(key)
        etag = self.processor.generate_etag()
        state = {"etag": etag}
        val = self.convert_put_value(value)
        expiry_time = (
            datetime.now(timezone.utc).timestamp() + (expiry / 1000)
            if expiry
            else 0
        )
        current_time = datetime.now(timezone.utc).timestamp()
        if where_exists is None and where_etag is None:
            query = f"""INSERT INTO {self.table} (
                            collection, id, value, etag, expiry
                        )
                        VALUES (%s, %s, %s, %s, %s)
                        ON CONFLICT (collection, id)
                        DO UPDATE SET
                        value = EXCLUDED.value,
                        etag = EXCLUDED.etag,
                        expiry = EXCLUDED.expiry
                    """
            params: tuple = (
                collection,
                id,
                val,
                etag,
                expiry_time,
            )
            return {
                "query": query,
                "params": params,
                "rowcount": True,
            }, state
        elif where_exists is False:
            query = f"""INSERT INTO {self.table} (
                            collection, id, value, etag, expiry
                        )
                        VALUES (%s, %s, %s, %s, %s)"""
            params = (
                collection,
                id,
                val,
                etag,
                expiry_time,
            )
            return {
                "query": query,
                "params": params,
                "rowcount": True,
            }, state
        elif where_exists is True:
            query = f"""
                    UPDATE {self.table}
                    SET value = %s,
                    etag = %s,
                    expiry = %s
                    WHERE collection = %s
                    AND id = %s
                    AND (expiry = 0 OR expiry > %s)
                    """
            params = (
                val,
                etag,
                expiry_time,
                collection,
                id,
                current_time,
            )
            return {
                "query": query,
                "params": params,
                "rowcount": True,
            }, state
        else:
            query = f"""
                    UPDATE {self.table}
                    SET value = %s,
                    etag = %s,
                    expiry = %s
                    WHERE collection = %s
                    AND id = %s
                    AND etag = %s
                    AND (expiry = 0 OR expiry > %s)
                    """
            params = (
                val,
                etag,
                expiry_time,
                collection,
                id,
                where_etag,
                current_time,
            )
            return {
                "query": query,
                "params": params,
                "rowcount": True,
            }, state

    def convert_update(
        self,
        key: KeyValueKeyType,
        set: Update,
        where_etag: str | None,
        returning: str | None,
        collection: str | None,
    ) -> tuple[dict, dict]:
        id = self.processor.get_id_from_key(key)
        etag = self.processor.generate_etag()
        state = {"etag": etag}
        not_supported = False
        expiry_time = 0
        if len(set.operations) == 1 and (
            set.operations[0].field == Attribute.VALUE
            or set.operations[0].field == UpdateAttribute.VALUE
        ):
            op = set.operations[0].op
            arg = set.operations[0].args[0]
            query = None
            if op == UpdateOp.INCREMENT:
                query = f"""
                        INSERT INTO {self.table}
                        (collection, id, value, etag, expiry)
                        VALUES (%s, %s, (%s::TEXT)::BYTEA, %s, %s)
                        ON CONFLICT (collection, id) DO UPDATE
                        SET value =
                        ((convert_from(
                        {self.table}.value, 'UTF8')::NUMERIC +
                        convert_from(
                        EXCLUDED.value, 'UTF8')::NUMERIC
                        )::TEXT)::BYTEA,
                        etag = EXCLUDED.etag
                        """
            elif op == UpdateOp.APPEND:
                query = f"""
                        INSERT INTO {self.table}
                        (collection, id, value, etag, expiry)
                        VALUES (%s, %s, %s::BYTEA, %s, %s)
                        ON CONFLICT(collection, id) DO UPDATE
                        SET value =
                        (convert_from(
                        {self.table}.value, 'UTF8')
                        || convert_from(
                        EXCLUDED.value, 'UTF8'))::BYTEA,
                        etag = EXCLUDED.etag
                        """
            elif op == UpdateOp.PREPEND:
                query = f"""
                        INSERT INTO {self.table}
                        (collection, id, value, etag, expiry)
                        VALUES (%s, %s, %s::BYTEA, $s, %s)
                        ON CONFLICT(collection, id) DO UPDATE
                        SET value =
                        (convert_from(
                        EXCLUDED.value, 'UTF8')
                        || convert_from(
                        {self.table}.value, 'UTF8'))::BYTEA,
                        etag = EXCLUDED.etag
                        """
            else:
                not_supported = True
        else:
            not_supported = True
        if not_supported:
            raise BadRequestError(f"Update operation not supported {set}")
        params: tuple = (
            collection,
            id,
            arg,
            etag,
            expiry_time,
        )
        if where_etag:
            query = f"{query} WHERE {self.table}.etag = %s"
            params = params + (where_etag,)
        if returning == "new":
            query = f"{query} RETURNING value"
        else:
            query = f"{query} RETURNING id"
        return {
            "query": query,
            "params": params,
            "rowcount": not returning,
        }, state

    def convert_delete(
        self,
        key: KeyValueKeyType,
        where_etag: str | None,
        collection: str | None,
    ) -> dict:
        id = self.processor.get_id_from_key(key)
        query = f"""
                    DELETE FROM {self.table}
                    WHERE collection = %s AND id = %s
                    """
        if where_etag:
            query = f"{query} AND etag = %s"
            params: tuple = (
                collection,
                id,
                where_etag,
            )
        else:
            params = (
                collection,
                id,
            )
        query = f"{query} RETURNING expiry"
        return {
            "query": query,
            "params": params,
        }

    def convert_query(
        self,
        where: Expression | None,
        limit: int | None,
        collection: str | None,
    ) -> dict:
        where_str = self._convert_where(where, collection)
        query = f"SELECT id FROM {self.table} WHERE {where_str}"
        if limit:
            query = f"{query} LIMIT {limit}"
        return {"query": query, "fetchall": True}

    def convert_count(
        self,
        where: Expression | None,
        collection: str | None,
    ) -> dict:
        where_str = self._convert_where(where, collection)
        query = f"SELECT COUNT(*) FROM {self.table} WHERE {where_str}"
        return {"query": query}

    def convert_batch(
        self,
        op_parsers: list[StoreOperationParser],
        collection: str | None,
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for op_parser in op_parsers:
            collection = op_parser.get_collection_name() or collection
            id = self.processor.get_id_from_key(op_parser.get_key())
            if op_parser.op_equals(StoreOperation.PUT):
                etag = self.processor.generate_etag()
                query = f"""INSERT INTO {self.table} (
                            collection, id, value, etag, expiry
                        )
                        VALUES (%s, %s, %s, %s, %s)
                        ON CONFLICT (collection, id)
                        DO UPDATE SET
                        value = EXCLUDED.value,
                        etag = EXCLUDED.etag,
                        expiry = EXCLUDED.expiry
                        """
                params: tuple = (
                    collection,
                    id,
                    op_parser.get_value(),
                    etag,
                    0,
                )
                ops.append(
                    {"query": query, "params": params, "rowcount": True}
                )
                states.append({"etag": etag})
            elif op_parser.op_equals(StoreOperation.GET):
                query = f"""
                        SELECT id, value, etag, expiry
                        FROM {self.table}
                        WHERE collection = %s AND id = %s
                        """
                params = (
                    collection,
                    id,
                )
                ops.append({"query": query, "params": params})
                states.append(None)
            elif op_parser.op_equals(StoreOperation.DELETE):
                query = f"""
                        DELETE FROM {self.table}
                        WHERE collection = %s AND id = %s
                        """
                params = (
                    collection,
                    id,
                )
                ops.append(
                    {"query": query, "params": params, "rowcount": True}
                )
                states.append(None)
            else:
                raise BadRequestError(
                    f"Operation not supported in batch: {op_parser}"
                )
        return {"ops": ops}, {"etags": states}

    def convert_evict(
        self,
        key: KeyValueKeyType,
        collection: str | None,
    ) -> dict:
        id = self.processor.get_id_from_key(key)
        current_time = datetime.now(timezone.utc).timestamp()
        query = f"""
                DELETE FROM {self.table}
                WHERE collection = %s AND id = %s
                AND expiry != 0
                AND expiry < %s
                """
        return {
            "query": query,
            "params": (
                collection,
                id,
                current_time,
            ),
        }

    def _convert_where(
        self, where: Expression | None, collection: str | None
    ) -> str:
        if collection:
            query = f"collection = '{collection}'"
        else:
            query = "collection = null"
        if where is None:
            return query
        if isinstance(where, Comparison):
            if where.op == ComparisonOp.LIKE and isinstance(where.rexpr, str):
                return f"{query} AND id LIKE {where.rexpr}"
        elif isinstance(where, Function):
            if where.namespace == FunctionNamespace.BUILTIN:
                if where.name == QueryFunctionName.STARTS_WITH:
                    return f"{query} AND id LIKE '{where.args[1]}%'"
                if where.name == QueryFunctionName.ENDS_WITH:
                    return f"{query} AND id LIKE '%{where.args[1]}'"
                if where.name == QueryFunctionName.CONTAINS:
                    return f"{query} AND id LIKE '%{where.args[1]}%'"
        raise BadRequestError(f"Condition not supported: {where!r}")


class ResultConverter:
    processor: ItemProcessor

    def __init__(
        self,
        processor: ItemProcessor,
    ):
        self.processor = processor

    def convert_get(self, nresult: Any, type: str) -> KeyValueItem:
        (id, value, etag, expiry) = nresult
        return build_item(id=id, value=convert_value(value, type), etag=etag)

    def convert_put(
        self,
        nresult: Any,
        key: KeyValueKeyType,
        value: KeyValueValueType,
        where_exists: bool | None,
        where_etag: str | None,
        returning: str | None,
        etag: str,
        type: str,
    ) -> KeyValueItem:
        id = self.processor.get_id_from_key(key)
        if where_exists is True or where_etag:
            if nresult == 0:
                raise PreconditionFailedError
        return_value = None
        if returning == "new":
            return_value = value
        return build_item(
            id=id, value=convert_value(return_value, type), etag=etag
        )

    def convert_update(
        self,
        nresult: Any,
        key: KeyValueKeyType,
        where_etag: str | None,
        returning: str | None,
        etag: str,
        type: str,
    ) -> KeyValueItem:
        id = self.processor.get_id_from_key(key)
        return_value = None
        if nresult is None:
            raise PreconditionFailedError
        if returning == "new":
            (return_value,) = nresult
        return build_item(
            id=id, value=convert_value(return_value, type), etag=etag
        )

    def convert_delete(self, nresult: Any, where_etag: str | None):
        if nresult is None:
            if where_etag:
                raise PreconditionFailedError
            raise NotFoundError
        current_time = datetime.now(timezone.utc).timestamp()
        if nresult[0] != 0 and current_time > nresult[0]:
            raise NotFoundError
        return None

    def convert_query(self, nresult: Any) -> KeyValueList:
        items: list = []
        for raw_item in nresult:
            items.append(build_item(id=raw_item[0]))
        return KeyValueList(items=items)

    def convert_count(self, nresult: Any) -> int:
        return nresult[0]

    def convert_batch(
        self,
        nresult: Any,
        op_parsers: list[StoreOperationParser],
        states: list[dict],
        type: str,
    ) -> list[Any]:
        result: list = []
        for op_parser, op_result, state in zip(op_parsers, nresult, states):
            id = self.processor.get_id_from_key(op_parser.get_key())
            if op_parser.op_equals(StoreOperation.PUT):
                result.append(build_item(id=id, etag=state["etag"]))
            elif op_parser.op_equals(StoreOperation.GET):
                expiry = op_result[3]
                current_time = datetime.now(timezone.utc).timestamp()
                has_expired = expiry != 0 and current_time > expiry
                if has_expired:
                    result.append(None)
                else:
                    result.append(
                        build_item(
                            id=id,
                            value=convert_value(op_result[1], type),
                            etag=op_result[2],
                        )
                    )
            elif op_parser.op_equals(StoreOperation.DELETE):
                result.append(None)
        return result


class ClientHelper:
    client: Any
    op_converter: OperationConverter

    def __init__(self, client: Any, op_converter: OperationConverter):
        self.client = client
        self.op_converter = op_converter

    def exists(
        self,
        key: KeyValueKeyType,
        collection: str | None,
    ) -> bool:
        args = self.op_converter.convert_exists(key, collection)
        cursor = self.client.cursor()
        try:
            cursor.execute(args["query"], args["params"])
            nresult = cursor.fetchone()
            if nresult is None:
                return False
            (id, expiry) = nresult
            current_time = datetime.now(timezone.utc).timestamp()
            has_expired = expiry != 0 and current_time > expiry
            if has_expired:
                args = self.op_converter.convert_evict(key, collection)
                cursor.execute(args["query"], args["params"])
                return False
        finally:
            self.client.commit()
            cursor.close()
        return True

    def get(
        self,
        key: KeyValueKeyType,
        collection: str | None,
        start: int | None,
        end: int | None,
    ) -> Any:
        args = self.op_converter.convert_get(key, start, end, collection)
        cursor = self.client.cursor()
        try:
            cursor.execute(args["query"], args["params"])
            nresult = cursor.fetchone()
            if nresult is None:
                raise NotFoundError
            (
                id,
                value,
                etag,
                expiry,
            ) = nresult
            current_time = datetime.now(timezone.utc).timestamp()
            has_expired = expiry != 0 and current_time > expiry
            if has_expired:
                args = self.op_converter.convert_evict(key, collection)
                cursor.execute(args["query"], args["params"])
                raise NotFoundError
        finally:
            self.client.commit()
            cursor.close()
        return nresult

    def execute(
        self,
        query: str,
        params: tuple | None = None,
        rowcount: bool = False,
        fetchall: bool = False,
    ) -> Any:
        cursor = self.client.cursor()
        try:
            if params:
                cursor.execute(query, params)
            else:
                cursor.execute(query)
            if rowcount:
                return cursor.rowcount
            if fetchall:
                return cursor.fetchall()
            if cursor.rowcount > 0:
                return cursor.fetchone()
        finally:
            self.client.commit()
            cursor.close()

    def batch(self, ops: list) -> list[Any]:
        nresult: list = []
        cursor = self.client.cursor()
        try:
            for op in ops:
                if "params" in op:
                    cursor.execute(op["query"], op["params"])
                else:
                    cursor.execute(op["query"])
                if "rowcount" in op and op["rowcount"]:
                    nresult.append(None)
                else:
                    r = cursor.fetchone()
                    nresult.append(r)
        finally:
            self.client.commit()
            cursor.close()
        return nresult

    def close(self) -> Any:
        self.client.close()


class AsyncClientHelper:
    client: Any
    op_converter: OperationConverter

    def __init__(self, client: Any, op_converter: OperationConverter):
        self.client = client
        self.op_converter = op_converter

    async def exists(
        self,
        key: KeyValueKeyType,
        collection: str | None,
    ) -> bool:
        args = self.op_converter.convert_exists(key, collection)
        cursor = self.client.cursor()
        try:
            await cursor.execute(args["query"], args["params"])
            nresult = await cursor.fetchone()
            if nresult is None:
                return False
            (id, expiry) = nresult
            current_time = datetime.now(timezone.utc).timestamp()
            has_expired = expiry != 0 and current_time > expiry
            if has_expired:
                args = self.op_converter.convert_evict(key, collection)
                await cursor.execute(args["query"], args["params"])
                return False
        finally:
            await self.client.commit()
            await cursor.close()
        return True

    async def get(
        self,
        key: KeyValueKeyType,
        collection: str | None,
        start: int | None,
        end: int | None,
    ) -> Any:
        args = self.op_converter.convert_get(key, start, end, collection)
        cursor = self.client.cursor()
        try:
            await cursor.execute(args["query"], args["params"])
            nresult = await cursor.fetchone()
            if nresult is None:
                raise NotFoundError
            (
                id,
                value,
                etag,
                expiry,
            ) = nresult
            current_time = datetime.now(timezone.utc).timestamp()
            has_expired = expiry != 0 and current_time > expiry
            if has_expired:
                args = self.op_converter.convert_evict(key, collection)
                await cursor.execute(args["query"], args["params"])
                raise NotFoundError
        finally:
            await self.client.commit()
            await cursor.close()
        return nresult

    async def execute(
        self,
        query: str,
        params: tuple | None = None,
        rowcount: bool = False,
        fetchall: bool = False,
    ) -> Any:
        cursor = self.client.cursor()
        try:
            if params:
                await cursor.execute(query, params)
            else:
                await cursor.execute(query)
            if rowcount:
                return cursor.rowcount
            if fetchall:
                return await cursor.fetchall()
            if cursor.rowcount > 0:
                return await cursor.fetchone()
        finally:
            await self.client.commit()
            await cursor.close()

    async def batch(self, ops: list) -> list[Any]:
        nresult: list = []
        cursor = self.client.cursor()
        try:
            for op in ops:
                if "params" in op:
                    await cursor.execute(op["query"], op["params"])
                else:
                    await cursor.execute(op["query"])
                if "rowcount" in op and op["rowcount"]:
                    nresult.append(None)
                else:
                    r = await cursor.fetchone()
                    nresult.append(r)
        finally:
            await self.client.commit()
            await cursor.close()
        return nresult

    async def close(self) -> Any:
        await self.client.close()



"""
Key Value Store on Redis.
"""

from __future__ import annotations

__all__ = ["Redis"]

from typing import Any

from x8._common.redis_provider import RedisProvider
from x8.core import Context, NCall, Operation, Response
from x8.core.exceptions import (
    BadRequestError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import (
    Comparison,
    ComparisonOp,
    Expression,
    Function,
    FunctionNamespace,
    QueryFunctionName,
    Update,
    UpdateOp,
)
from x8.storage._common import (
    Attribute,
    ItemProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    UpdateAttribute,
)

from .._helper import (
    build_item,
    convert_value,
    get_collection_key,
    get_collection_name,
    get_collection_prefix,
    get_query_config,
)
from .._models import (
    KeyValueItem,
    KeyValueKeyType,
    KeyValueList,
    KeyValueQueryConfig,
    KeyValueValueType,
)


class Redis(RedisProvider, StoreProvider):
    collection: str | None
    nparams: dict[str, Any]

    _lib: Any
    _alib: Any

    _client: Any
    _aclient: Any
    _processor: ItemProcessor
    _op_converter: OperationConverter
    _result_converter: ResultConverter

    def __init__(
        self,
        url: str | None = None,
        host: str | None = None,
        port: int | None = None,
        db: int = 0,
        username: str | None = None,
        password: str | None = None,
        options: dict | None = None,
        collection: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            url:
                Redis endpoint url.
            host:
                Redis endpoint host name.
            port:
                Redis endpoint port.
            db:
                Redis database number.
            username:
                Username for auth.
            password:
                Password for auth.
            options:
                Redis client options.
            collection:
                Collection name.
            nparams:
                Native parameters to the redis client.
        """
        self.collection = collection
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._processor = ItemProcessor()
        self._op_converter = OperationConverter(processor=self._processor)
        self._result_converter = ResultConverter(processor=self._processor)
        RedisProvider.__init__(
            self,
            url=url,
            host=host,
            port=port,
            db=db,
            username=username,
            password=password,
            options=options,
            **kwargs,
        )

    def __setup__(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        decode_responses = self.__component__.type != "binary"
        self._client, self._lib = self._get_client_and_lib(
            decode_responses=decode_responses,
        )

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        decode_responses = self.__component__.type != "binary"
        self._aclient, self._alib = self._aget_client_and_lib(
            decode_responses=decode_responses,
        )

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = self.get_op_parser(operation)
        ncall, state = self._get_ncall(
            self._client,
            op_parser,
            ClientHelper(
                self._client,
                self._lib,
                self._op_converter,
            ),
        )
        if ncall is None:
            return super().__run__(
                operation,
                context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.__asetup__(context=context)
        op_parser = self.get_op_parser(operation)
        ncall, state = self._get_ncall(
            self._aclient,
            op_parser,
            AsyncClientHelper(
                self._aclient,
                self._alib,
                self._op_converter,
            ),
        )
        if ncall is None:
            return await super().__arun__(
                operation,
                context,
                **kwargs,
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        client: Any,
        op_parser: StoreOperationParser,
        client_helper: Any,
    ) -> tuple[NCall | None, dict]:
        call = None
        state: dict = {}
        nargs = op_parser.get_nargs()
        collection: str | None = get_collection_name(self, op_parser)
        op_converter = self._op_converter
        # EXISTS
        if op_parser.op_equals(StoreOperation.EXISTS):
            args: Any = op_converter.convert_exists(
                op_parser.get_key(),
                collection,
            )
            call = NCall(client.exists, args, nargs)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = op_converter.convert_get(
                op_parser.get_key(),
                op_parser.get_start(),
                op_parser.get_end(),
                collection,
            )
            call = NCall(client.hmget, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            etag = self._processor.generate_etag()
            args = {
                "key": op_parser.get_key(),
                "value": op_parser.get_value(),
                "where_exists": op_parser.get_where_exists(),
                "where_etag": op_parser.get_where_etag(),
                "expiry": op_parser.get_expiry(),
                "returning": op_parser.get_returning(),
                "collection": collection,
                "etag": etag,
                "nargs": nargs,
            }
            state["etag"] = etag
            call = NCall(
                client_helper.put,
                args,
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            etag = self._processor.generate_etag()
            args = {
                "key": op_parser.get_key(),
                "set": op_parser.get_set(),
                "where_exists": op_parser.get_where_exists(),
                "where_etag": op_parser.get_where_etag(),
                "returning": op_parser.get_returning(),
                "collection": collection,
                "etag": etag,
                "nargs": nargs,
            }
            state["etag"] = etag
            call = NCall(client_helper.update, args)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = {
                "key": op_parser.get_key(),
                "where_exists": op_parser.get_where_exists(),
                "where_etag": op_parser.get_where_etag(),
                "collection": collection,
                "nargs": nargs,
            }
            call = NCall(client_helper.delete, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = op_converter.convert_query(
                op_parser.get_where(),
                op_parser.get_limit(),
                op_parser.get_continuation(),
                get_query_config(op_parser),
                collection,
            )
            args["nargs"] = nargs
            call = NCall(client_helper.query, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = op_converter.convert_count(
                op_parser.get_where(),
                collection,
            )
            args["nargs"] = nargs
            call = NCall(client_helper.count, args)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args = {
                "op_parsers": op_parser.get_operation_parsers(),
                "collection": collection,
                "nargs": nargs,
            }
            call = NCall(client_helper.batch, args)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(client_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: StoreOperationParser,
    ) -> Any:
        result_converter = self._result_converter
        result: Any = None
        type = self.__component__.type
        # EXISTS
        if op_parser.op_equals(StoreOperation.EXISTS):
            result = result_converter.convert_exists(nresult)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            result = result_converter.convert_get(
                nresult,
                op_parser.get_key(),
                op_parser.get_start(),
                op_parser.get_end(),
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = result_converter.convert_put(
                nresult,
                op_parser.get_key(),
                op_parser.get_returning(),
                op_parser.get_value(),
                state["etag"],
                type,
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            result = result_converter.convert_update(
                nresult,
                op_parser.get_key(),
                op_parser.get_returning(),
                state["etag"],
                type,
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = result_converter.convert_delete(
                nresult,
                op_parser.get_where_exists(),
                op_parser.get_where_etag(),
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            result = result_converter.convert_query(nresult)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = result_converter.convert_count(nresult)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = nresult
        return result


class OperationConverter:
    processor: ItemProcessor

    def __init__(self, processor: ItemProcessor):
        self.processor = processor

    def convert_to_db_key(
        self, key: KeyValueKeyType, collection: str | None
    ) -> str:
        id = self.processor.get_id_from_key(key)
        db_key = get_collection_key(collection, id)
        return db_key

    def convert_exists(
        self,
        key: KeyValueKeyType,
        collection: str | None,
    ) -> list:
        db_key = self.convert_to_db_key(key, collection)
        return [db_key]

    def convert_get(
        self,
        key: KeyValueKeyType,
        start: int | None,
        end: int | None,
        collection: str | None,
    ) -> dict:
        db_key = self.convert_to_db_key(key, collection)
        args = {"name": db_key, "keys": ["value", "etag"]}
        return args

    def convert_put(
        self,
        key: KeyValueKeyType,
        value: KeyValueValueType,
        where_exists: bool | None,
        where_etag: str | None,
        expiry: int | None,
        returning: str | None,
        collection: str | None,
        etag: str,
    ) -> dict:
        db_key = self.convert_to_db_key(key, collection)
        args: dict = {
            "name": db_key,
            "mapping": {
                "value": value,
                "etag": etag,
            },
        }
        return args

    def convert_update(
        self,
        key: KeyValueKeyType,
        set: Update,
        where_exists: bool | None,
        where_etag: str | None,
        returning: str | None,
        collection: str | None,
        etag: str,
    ) -> tuple[dict, str]:
        db_key = self.convert_to_db_key(key, collection)
        if len(set.operations) == 1 and (
            set.operations[0].field == Attribute.VALUE
            or set.operations[0].field == UpdateAttribute.VALUE
        ):
            op = set.operations[0].op
            if op == UpdateOp.INCREMENT:
                amount = set.operations[0].args[0]
                args: dict = {
                    "name": db_key,
                    "key": "value",
                    "amount": amount,
                }
                if isinstance(amount, int):
                    return args, "hincrby"
                else:
                    return args, "hincrbyfloat"
        raise BadRequestError(f"Update operation not supported: {set}")

    def convert_delete(
        self,
        key: KeyValueKeyType,
        where_exists: bool | None,
        where_etag: str | None,
        collection: str | None,
    ) -> list:
        db_key = self.convert_to_db_key(key, collection)
        return [db_key]

    def convert_query(
        self,
        where: Expression | None,
        limit: int | None,
        continuation: str | None,
        config: KeyValueQueryConfig | None,
        collection: str | None,
    ) -> dict:
        args: dict = {}
        args["continuation"] = continuation
        args["limit"] = limit
        args["match"] = self._convert_where(where, collection)
        args["config"] = config
        args["collection"] = collection
        return args

    def convert_count(
        self,
        where: Expression | None,
        collection: str | None,
    ) -> dict:
        if where is None and collection is None:
            return {"func": "dbsize", "match": None}
        pattern = self._convert_where(where, collection)
        return {"func": "scan_iter", "match": pattern}

    def convert_batch(
        self,
        op_parsers: list[StoreOperationParser],
        collection: str | None,
    ) -> tuple[list, str]:
        def is_batch_op(
            op_parsers: list[StoreOperationParser], op: str
        ) -> bool:
            for op_parser in op_parsers:
                if not op_parser.op_equals(op):
                    return False
            return True

        args_list: list = []
        if is_batch_op(op_parsers, StoreOperation.PUT):
            for op_parser in op_parsers:
                col = op_parser.get_collection_name() or collection
                db_key = self.convert_to_db_key(op_parser.get_key(), col)
                value = op_parser.get_value()
                etag = self.processor.generate_etag()
                args_list.append(
                    {
                        "name": db_key,
                        "mapping": {
                            "value": value,
                            "etag": etag,
                        },
                    }
                )
            return args_list, "hset"
        elif is_batch_op(op_parsers, StoreOperation.GET):
            for op_parser in op_parsers:
                col = op_parser.get_collection_name() or collection
                db_key = self.convert_to_db_key(op_parser.get_key(), col)
                args_list.append({"name": db_key, "keys": ["value", "etag"]})
            return args_list, "hmget"
        elif is_batch_op(op_parsers, StoreOperation.DELETE):
            for op_parser in op_parsers:
                col = op_parser.get_collection_name() or collection
                db_key = self.convert_to_db_key(op_parser.get_key(), col)
                args_list.append(db_key)
            return args_list, "delete"
        else:
            raise BadRequestError("Batch operations not supported")

    def _convert_where(
        self,
        where: Expression | None,
        collection: str | None,
    ) -> str | None:
        prefix = get_collection_prefix(collection)
        if where is None:
            return f"{prefix}*"
        if isinstance(where, Comparison):
            if where.op == ComparisonOp.LIKE and isinstance(where.rexpr, str):
                if prefix:
                    return f"{prefix}{where.rexpr}"
                return where.rexpr
        elif isinstance(where, Function):
            if where.namespace == FunctionNamespace.BUILTIN:
                if where.name == QueryFunctionName.STARTS_WITH:
                    if prefix:
                        return f"{prefix}{where.args[1]}*"
                    return where.args[1]
                if where.name == QueryFunctionName.ENDS_WITH:
                    if prefix:
                        return f"{prefix}*{where.args[1]}"
                    return where.args[1]
                if where.name == QueryFunctionName.CONTAINS:
                    if prefix:
                        return f"{prefix}*{where.args[1]}*"
                    return where.args[1]
        raise BadRequestError(f"Condition not supported: {where!r}")


class ResultConverter:
    processor: ItemProcessor

    def __init__(self, processor: ItemProcessor):
        self.processor = processor

    def convert_exists(self, nresult: Any) -> bool:
        if nresult == 1:
            return True
        return False

    def convert_get(
        self,
        nresult: Any,
        key: KeyValueKeyType,
        start: int | None,
        end: int | None,
    ) -> KeyValueItem:
        id = self.processor.get_id_from_key(key)
        if not nresult or not nresult[1]:
            raise NotFoundError
        value = nresult[0]
        if start and end:
            eend = end + 1
            value = value[start:eend]
        elif start:
            value = value[start:]
        elif end:
            eend = end + 1
            value = value[:eend]
        return build_item(id=id, value=value, etag=nresult[1])

    def convert_put(
        self,
        nresult: Any,
        key: KeyValueKeyType,
        returning: str | None,
        value: KeyValueValueType,
        etag: str,
        type: str,
    ) -> KeyValueItem:
        id = self.processor.get_id_from_key(key)
        if not nresult:
            raise PreconditionFailedError
        elif returning == "new":
            return build_item(id, convert_value(value, type), etag=etag)
        return build_item(id, etag=etag)

    def convert_update(
        self,
        nresult: Any,
        key: KeyValueKeyType,
        returning: str | None,
        etag: str,
        type: str,
    ) -> KeyValueItem:
        if nresult is None:
            raise PreconditionFailedError
        return_value = None
        if returning == "new":
            return_value = nresult
        id = self.processor.get_id_from_key(key)
        return build_item(id, convert_value(return_value, type), etag=etag)

    def convert_delete(
        self,
        nresult: Any,
        where_exists: bool | None,
        where_etag: str | None,
    ):
        if nresult == 0:
            if where_exists is not None or where_etag:
                raise PreconditionFailedError
            raise NotFoundError
        return None

    def convert_query(self, nresult: Any):
        return nresult

    def convert_count(self, nresult: Any):
        return nresult


class ClientHelper:
    client: Any
    lib: Any
    op_converter: OperationConverter

    def __init__(
        self,
        client: Any,
        lib: Any,
        op_converter: OperationConverter,
    ):
        self.client = client
        self.lib = lib
        self.op_converter = op_converter

    def put(
        self,
        key: KeyValueKeyType,
        value: KeyValueValueType,
        where_exists: bool | None,
        where_etag: str | None,
        expiry: int | None,
        returning: str | None,
        collection: str | None,
        etag: str,
        nargs: Any,
    ) -> Any:
        args = self.op_converter.convert_put(
            key,
            value,
            where_exists,
            where_etag,
            expiry,
            returning,
            collection,
            etag,
        )
        if expiry is None and where_exists is None and where_etag is None:
            NCall(self.client.hset, args, nargs).invoke()
            return True
        db_key = self.op_converter.convert_to_db_key(key, collection)
        with self.client.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(db_key)
                    if where_exists is True:
                        if not pipe.exists(db_key):
                            pipe.unwatch()
                            return False
                    elif where_exists is False:
                        if pipe.exists(db_key):
                            return False
                    elif where_etag:
                        etag = pipe.hget(db_key, "etag")
                        if isinstance(etag, bytes):
                            etag = etag.decode()
                        if not etag or etag != where_etag:
                            return False
                    pipe.multi()
                    pipe.hset(**args)
                    if expiry:
                        pipe.pexpire(db_key, expiry)
                    pipe.execute()
                    return True
                except self.lib.WatchError:
                    continue

    def update(
        self,
        key: KeyValueKeyType,
        set: Update,
        where_exists: bool | None,
        where_etag: str | None,
        returning: str | None,
        collection: str | None,
        etag: str,
        nargs: Any,
    ) -> Any:
        args, func = self.op_converter.convert_update(
            key,
            set,
            where_exists,
            where_etag,
            returning,
            collection,
            etag,
        )
        db_key = self.op_converter.convert_to_db_key(key, collection)
        with self.client.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(db_key)
                    if where_etag:
                        etag = pipe.hget(db_key, "etag")
                        if isinstance(etag, bytes):
                            etag = etag.decode()
                        if not etag or etag != where_etag:
                            return None
                    pipe.multi()
                    if func == "hincrby":
                        pipe.hincrby(**args)
                    elif func == "hincrbyfloat":
                        pipe.hincrbyfloat(**args)
                    pipe.hset(db_key, "etag", etag)
                    responses = pipe.execute()
                    return responses[0]
                except self.lib.WatchError:
                    continue

    def delete(
        self,
        key: KeyValueKeyType,
        where_exists: bool | None,
        where_etag: str | None,
        collection: str | None,
        nargs: Any,
    ) -> Any:
        args = self.op_converter.convert_delete(
            key, where_exists, where_etag, collection
        )
        if where_exists is None and where_etag is None:
            return NCall(self.client.delete, args).invoke()
        db_key = self.op_converter.convert_to_db_key(key, collection)
        with self.client.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(db_key)
                    if where_etag:
                        etag = pipe.hget(db_key, "etag")
                        if isinstance(etag, bytes):
                            etag = etag.decode()
                        if not etag or etag != where_etag:
                            return False
                        pipe.delete(db_key)
                        return 1
                except self.lib.WatchError:
                    continue

    def query(
        self,
        continuation: str | None,
        limit: int | None,
        match: str | None,
        config: KeyValueQueryConfig | None,
        collection: str | None,
        nargs: Any,
    ) -> KeyValueList:
        items: list = []
        prefix_length = len(collection) + 1 if collection else 0
        rcontinuation = continuation
        rlimit = limit
        if config and config.paging and config.page_size:
            if limit:
                if config.page_size < limit:
                    rlimit = config.page_size
            else:
                rlimit = config.page_size
        while True:
            cursor = int(rcontinuation) if rcontinuation else 0
            count = rlimit if rlimit else None
            cursor, keys = self.client.scan(
                cursor=cursor, count=count, match=match
            )
            for key in keys:
                items.append(build_item(id=key[prefix_length:]))
            if cursor == 0:
                rcontinuation = None
                break
            if config and config.paging and not config.page_size:
                rcontinuation = str(cursor)
                break
            if rlimit and len(items) < rlimit:
                rlimit = rlimit - len(items)
            rcontinuation = str(cursor)
        return KeyValueList(items=items, continuation=rcontinuation)

    def count(self, func: str, match: str | None, nargs: Any) -> int:
        count = 0
        if func == "dbsize":
            count = self.client.dbsize()
        elif func == "scan_iter":
            count = sum(1 for _ in self.client.scan_iter(match))
        return count

    def batch(
        self,
        op_parsers: list[StoreOperationParser],
        collection: str | None,
        nargs: Any,
    ):
        args_list, func = self.op_converter.convert_batch(
            op_parsers, collection
        )
        result: list = []
        if func == "hset":
            with self.client.pipeline() as pipe:
                pipe.multi()
                for arg, op_parser in zip(args_list, op_parsers):
                    pipe.hset(**arg)
                    result.append(
                        build_item(
                            id=self.op_converter.processor.get_id_from_key(
                                op_parser.get_key()
                            ),
                            etag=arg["mapping"]["etag"],
                        )
                    )
                pipe.execute()
        elif func == "hmget":
            with self.client.pipeline() as pipe:
                pipe.multi()
                for arg in args_list:
                    pipe.hmget(**arg)
                responses = pipe.execute()
            for op_parser, response in zip(op_parsers, responses):
                response.append(
                    build_item(
                        id=self.op_converter.processor.get_id_from_key(
                            op_parser.get_key()
                        ),
                        value=response[0],
                        etag=response[1],
                    )
                )
        elif func == "delete":
            NCall(self.client.delete, args_list).invoke()
            result = [None] * len(args_list)
        return result

    def close(self, nargs: Any) -> Any:
        pass


class AsyncClientHelper:
    client: Any
    lib: Any
    op_converter: OperationConverter

    def __init__(
        self,
        client: Any,
        lib: Any,
        op_converter: OperationConverter,
    ):
        self.client = client
        self.lib = lib
        self.op_converter = op_converter

    async def put(
        self,
        key: KeyValueKeyType,
        value: KeyValueValueType,
        where_exists: bool | None,
        where_etag: str | None,
        expiry: int | None,
        returning: str | None,
        collection: str | None,
        etag: str,
        nargs: Any,
    ) -> Any:
        args = self.op_converter.convert_put(
            key,
            value,
            where_exists,
            where_etag,
            expiry,
            returning,
            collection,
            etag,
        )
        db_key = self.op_converter.convert_to_db_key(key, collection)
        if expiry is None and where_exists is None and where_etag is None:
            await NCall(self.client.hset, args, nargs).ainvoke()
            return True

        async with self.client.pipeline(transaction=True) as pipe:
            while True:
                try:
                    await pipe.watch(db_key)
                    if where_exists is True:
                        if not await pipe.exists(db_key):
                            await pipe.unwatch()
                            return False
                    elif where_exists is False:
                        if await pipe.exists(db_key):
                            await pipe.unwatch()
                            return False
                    elif where_etag:
                        etag = await pipe.hget(db_key, "etag")
                        if isinstance(etag, bytes):
                            etag = etag.decode()
                        if not etag or etag != where_etag:
                            await pipe.unwatch()
                            return False
                    pipe.multi()
                    await pipe.hset(**args)
                    if expiry:
                        await pipe.pexpire(db_key, expiry)
                    await pipe.execute()
                    return True
                except self.lib.WatchError:
                    continue

    async def update(
        self,
        key: KeyValueKeyType,
        set: Update,
        where_exists: bool | None,
        where_etag: str | None,
        returning: str | None,
        collection: str | None,
        etag: str,
        nargs: Any,
    ) -> Any:
        args, func = self.op_converter.convert_update(
            key,
            set,
            where_exists,
            where_etag,
            returning,
            collection,
            etag,
        )
        db_key = self.op_converter.convert_to_db_key(key, collection)
        async with self.client.pipeline(transaction=True) as pipe:
            while True:
                try:
                    await pipe.watch(db_key)
                    if where_etag:
                        etag = await pipe.hget(db_key, "etag")
                        if isinstance(etag, bytes):
                            etag = etag.decode()
                        if not etag or etag != where_etag:
                            return None
                    pipe.multi()
                    if func == "hincrby":
                        await pipe.hincrby(**args)
                    elif func == "hincrbyfloat":
                        await pipe.hincrbyfloat(**args)
                    await pipe.hset(db_key, "etag", etag)
                    responses = await pipe.execute()
                    return responses[0]
                except self.lib.WatchError:
                    continue

    async def delete(
        self,
        key: KeyValueKeyType,
        where_exists: bool | None,
        where_etag: str | None,
        collection: str | None,
        nargs: Any,
    ) -> Any:
        args = self.op_converter.convert_delete(
            key, where_exists, where_etag, collection
        )
        if where_exists is None and where_etag is None:
            return await NCall(self.client.delete, args).ainvoke()
        db_key = self.op_converter.convert_to_db_key(key, collection)
        async with self.client.pipeline() as pipe:
            while True:
                try:
                    await pipe.watch(db_key)
                    if where_etag:
                        etag = await pipe.hget(db_key, "etag")
                        if isinstance(etag, bytes):
                            etag = etag.decode()
                        if not etag or etag != where_etag:
                            return False
                        await pipe.delete(db_key)
                        return 1
                except self.lib.WatchError:
                    continue

    async def query(
        self,
        continuation: str | None,
        limit: int | None,
        match: str | None,
        config: KeyValueQueryConfig | None,
        collection: str | None,
        nargs: Any,
    ) -> KeyValueList:
        items: list = []
        prefix_length = len(collection) + 1 if collection else 0
        rcontinuation = continuation
        rlimit = limit
        if config and config.paging and config.page_size:
            if limit:
                if config.page_size < limit:
                    rlimit = config.page_size
            else:
                rlimit = config.page_size
        while True:
            cursor = int(rcontinuation) if rcontinuation else 0
            count = rlimit if rlimit else None
            cursor, keys = await self.client.scan(
                cursor=cursor, count=count, match=match
            )
            for key in keys:
                items.append(build_item(id=key[prefix_length:]))
            if cursor == 0:
                rcontinuation = None
                break
            if config and config.paging and not config.page_size:
                rcontinuation = str(cursor)
                break
            if rlimit and len(items) < rlimit:
                rlimit = rlimit - len(items)
            rcontinuation = str(cursor)
        return KeyValueList(items=items, continuation=rcontinuation)

    async def count(
        self,
        func: str,
        match: str | None,
        nargs: Any,
    ):
        count = 0
        if func == "dbsize":
            count = await self.client.dbsize()
        elif func == "scan_iter":
            async for _ in self.client.scan_iter(match):
                count += 1
        return count

    async def batch(
        self,
        op_parsers: list[StoreOperationParser],
        collection: str | None,
        nargs: Any,
    ):
        args_list, func = self.op_converter.convert_batch(
            op_parsers, collection
        )
        result: list = []
        if func == "hset":
            async with self.client.pipeline() as pipe:
                for arg, op_parser in zip(args_list, op_parsers):
                    await pipe.hset(**arg)
                    result.append(
                        build_item(
                            id=self.op_converter.processor.get_id_from_key(
                                op_parser.get_key()
                            ),
                            etag=arg["mapping"]["etag"],
                        )
                    )
                await pipe.execute()
        elif func == "hmget":
            async with self.client.pipeline() as pipe:
                for arg in args_list:
                    await pipe.hmget(**arg)
                responses = await pipe.execute()
            for op_parser, response in zip(op_parsers, responses):
                result.append(
                    build_item(
                        id=self.op_converter.processor.get_id_from_key(
                            op_parser.get_key()
                        ),
                        value=response[0] if response else None,
                        etag=response[1] if len(response) > 1 else None,
                    )
                )

        elif func == "delete":
            await NCall(self.client.delete, args_list).ainvoke()
            result = [None] * len(args_list)

        return result

    async def close(self, nargs: Any) -> Any:
        await self.client.aclose()



messages = [
    dict(
        value="Test message 1",
        metadata={"str": "hello", "int": "10", "bool": "False"},
        properties=dict(
            message_id="id1",
            group_id="group1",
            content_type="text/plain",
        ),
    ),
    dict(
        value=b"Hello world",
        metadata={"str": "hello", "int": "10", "bool": "False"},
        properties=dict(
            message_id="id2",
            group_id="group2",
            content_type="application/octet-stream",
        ),
    ),
    dict(
        value={"str": "world", "int": 20, "bool": True},
        metadata={"str": "hello", "int": "10", "bool": "False"},
        properties=dict(
            message_id="id3",
            group_id="group2",
            content_type="application/json",
        ),
    ),
]
